# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional

from magika.types.content_type_info import ContentTypeInfo
from magika.types.magika_prediction import MagikaPrediction
from magika.types.status import Status


class MagikaResult:
    def __init__(
        self,
        *,
        path: Path,
        status: Status = Status.OK,
        prediction: Optional[MagikaPrediction] = None,
    ):
        self._path = path
        self._status = status
        self._prediction = prediction

    def __post_init__(self) -> None:
        assert self._path is not None
        if self._status == Status.OK:
            if self._prediction is None:
                raise ValueError("prediction must be set when status == OK")
        else:
            if self._prediction is not None:
                raise ValueError("prediction cannot be set when status != OK")

    @property
    def path(self) -> Path:
        return self._path

    @property
    def ok(self) -> bool:
        return self._status == Status.OK

    @property
    def status(self) -> Status:
        return self._status

    @property
    def prediction(self) -> MagikaPrediction:
        if self.ok:
            assert self._prediction is not None
            return self._prediction
        raise ValueError("prediction is not set when status != OK")

    # In the vast majority of cases, Magika API would return a status == OK (and
    # there is no code path that leads `identify_bytes()` to return an error).
    # To optimize for such frequent scenario, we add the following properties to
    # forward the underlying value. Clients that want to make use of the full
    # power of the absl-like StatusOr pattern can still do so, but we do not
    # force all clients, regardless of their complexity or criticality, to use
    # the more verbose `mr.prediction.output`.
    @property
    def dl(self) -> ContentTypeInfo:
        return self.prediction.dl

    @property
    def output(self) -> ContentTypeInfo:
        return self.prediction.output

    @property
    def score(self) -> float:
        return self.prediction.score

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if self.ok:
            return f"MagikaResult(path={self.path}, status={self.status}, prediction={self.prediction})"
        else:
            return f"MagikaResult(path={self.path}, status={self.status})"



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string
from pathlib import Path
from typing import List


def get_tests_data_dir() -> Path:
    repo_root_dir = Path(__file__).parent.parent
    tests_data_dir = repo_root_dir / "tests_data"
    if tests_data_dir.is_dir():
        return tests_data_dir
    tests_data_dir = repo_root_dir.parent / "tests_data"
    assert tests_data_dir.is_dir()
    return tests_data_dir


def get_basic_tests_files_dir() -> Path:
    tests_files_dir = get_tests_data_dir() / "basic"
    assert tests_files_dir.is_dir()
    return tests_files_dir


def get_mitra_tests_files_dir() -> Path:
    tests_files_dir = get_tests_data_dir() / "mitra"
    assert tests_files_dir.is_dir()
    return tests_files_dir


def get_previously_missdetected_files_dir() -> Path:
    tests_files_dir = get_tests_data_dir() / "previous_missdetections"
    assert tests_files_dir.is_dir()
    return tests_files_dir


def get_basic_test_files_paths() -> List[Path]:
    tests_files_dir = get_basic_tests_files_dir()
    test_files_paths = sorted(filter(lambda p: p.is_file(), tests_files_dir.rglob("*")))
    return test_files_paths


def get_mitra_test_files_paths() -> List[Path]:
    tests_files_dir = get_mitra_tests_files_dir()
    test_files_paths = sorted(filter(lambda p: p.is_file(), tests_files_dir.rglob("*")))
    return test_files_paths


def get_previously_missdetected_files_paths() -> List[Path]:
    tests_files_dir = get_previously_missdetected_files_dir()
    test_files_paths = sorted(filter(lambda p: p.is_file(), tests_files_dir.rglob("*")))
    return test_files_paths


def get_one_basic_test_file_path() -> Path:
    return get_basic_test_files_paths()[0]


def get_random_ascii_bytes(size: int) -> bytes:
    return bytes(
        [
            random.choice(bytes(string.ascii_letters.encode("ascii")))
            for _ in range(size)
        ]
    )


def get_lines_from_stream(stream: str) -> List[str]:
    candidates = stream.split("\n")
    lines = []
    for line in candidates:
        line = line.strip()
        if line == "":
            continue
        lines.append(line)
    return lines


def write_random_file_with_size(sample_path: Path, sample_size: int) -> None:
    print(f"Writing random file at {str(sample_path)} with size {sample_size}")
    assert not sample_path.is_file()
    block_size = 1024 * 1024 * 1024  # 1GB
    with open(sample_path, "wb") as f:
        if sample_size > block_size:
            for _ in range(sample_size // block_size):
                f.write(b"A" * block_size)
            if sample_size % block_size != 0:
                f.write(b"A" * (sample_size % block_size))
        else:
            f.write(b"A" * sample_size)
    print("Random file created")


def get_models_dir() -> Path:
    return Path(__file__).parent.parent / "src" / "magika" / "models"


def get_default_model_dir() -> Path:
    from magika.magika import Magika

    return get_models_dir() / Magika._get_default_model_name()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import signal
import tempfile
from pathlib import Path
from typing import Any, List, Optional

import pytest

from magika import Magika, PredictionMode
from magika.types import (
    ContentTypeInfo,
    ContentTypeLabel,
    MagikaPrediction,
    MagikaResult,
    Status,
)
from magika.types.overwrite_reason import OverwriteReason
from tests import utils


@pytest.mark.smoketest
def test_magika_module_check_version() -> None:
    import magika as magika_module

    assert isinstance(magika_module.__version__, str)

    m = Magika()
    assert m.get_module_version() == magika_module.__version__

    # Check that, when we don't specify `model_dir`, Magika uses the default
    # model.
    assert m.get_model_name() == m._get_default_model_name()


@pytest.mark.smoketest
def test_magika_module_with_one_test_file() -> None:
    test_path = utils.get_one_basic_test_file_path()

    m = Magika()

    _ = m.identify_path(test_path)
    _ = m.identify_paths([test_path])
    _ = m.identify_bytes(test_path.read_bytes())
    with open(test_path, "rb") as f:
        _ = m.identify_stream(f)


@pytest.mark.smoketest
def test_magika_module_with_explicit_model_dir() -> None:
    model_dir = utils.get_default_model_dir()
    test_path = utils.get_one_basic_test_file_path()

    m = Magika(model_dir=model_dir)

    _ = m.identify_path(test_path)
    _ = m.identify_paths([test_path])
    _ = m.identify_bytes(test_path.read_bytes())
    with open(test_path, "rb") as f:
        _ = m.identify_stream(f)


def test_magika_module_with_basic_tests_by_paths() -> None:
    tests_paths = utils.get_basic_test_files_paths()

    m = Magika()
    results = m.identify_paths(tests_paths)
    check_results_vs_expected_results(tests_paths, results)


def test_magika_module_with_basic_tests_by_path() -> None:
    tests_paths = utils.get_basic_test_files_paths()

    m = Magika()

    for test_path in tests_paths:
        result = m.identify_path(test_path)
        check_result_vs_expected_result(test_path, result)


def test_magika_module_with_basic_tests_by_bytes() -> None:
    tests_paths = utils.get_basic_test_files_paths()

    m = Magika()

    for test_path in tests_paths:
        content = test_path.read_bytes()
        result = m.identify_bytes(content)
        check_result_vs_expected_result(
            test_path, result, expected_result_path=Path("-")
        )


def test_magika_module_with_basic_tests_by_stream() -> None:
    tests_paths = utils.get_basic_test_files_paths()

    m = Magika()

    for test_path in tests_paths:
        with open(test_path, "rb") as f:
            result = m.identify_stream(f)
        check_result_vs_expected_result(
            test_path, result, expected_result_path=Path("-")
        )


def test_magika_module_with_all_models() -> None:
    tests_paths = utils.get_basic_test_files_paths()

    models_dir = utils.get_models_dir()
    for model_dir in models_dir.iterdir():
        m = Magika(model_dir=model_dir)
        for test_path in tests_paths:
            result = m.identify_path(test_path)
            check_result_vs_expected_result(test_path, result)


def test_magika_module_with_previously_missdetected_samples() -> None:
    model_dir = utils.get_default_model_dir()
    tests_paths = utils.get_previously_missdetected_files_paths()

    m = Magika(model_dir=model_dir)
    results = m.identify_paths(tests_paths)
    check_results_vs_expected_results(tests_paths, results)


def test_magika_module_with_empty_content() -> None:
    m = Magika()

    empty_content = b""

    res = m.identify_bytes(empty_content)
    assert res.ok
    assert res.path == Path("-")
    assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
    assert res.prediction.output.label == ContentTypeLabel.EMPTY
    assert res.prediction.score == 1.0

    with tempfile.TemporaryDirectory() as td:
        tf_path = Path(td) / "empty.dat"
        tf_path.write_bytes(empty_content)
        res = m.identify_path(tf_path)
        assert res.path == tf_path
        assert res.ok
        assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
        assert res.prediction.output.label == ContentTypeLabel.EMPTY
        assert res.prediction.score == 1.0

    res = m.identify_stream(io.BytesIO(b""))
    assert res.path == Path("-")
    assert res.ok
    assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
    assert res.prediction.output.label == ContentTypeLabel.EMPTY
    assert res.prediction.score == 1.0


def test_magika_module_with_short_content() -> None:
    m = Magika()

    text_content = b"asd"
    binary_content = b"\x80\x80\x80"

    for content, expected_ct_label in zip(
        [text_content, binary_content],
        [ContentTypeLabel.TXT, ContentTypeLabel.UNKNOWN],
    ):
        with tempfile.TemporaryDirectory() as td:
            # prediction via path
            tf_path = Path(td) / "file.txt"
            tf_path.write_bytes(content)
            res = m.identify_path(tf_path)
            assert res.path == tf_path
            assert res.ok
            assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
            assert res.prediction.output.label == expected_ct_label
            assert res.prediction.score == 1.0

            # prediction via bytes
            res = m.identify_bytes(content)
            assert res.path == Path("-")
            assert res.ok
            assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
            assert res.prediction.output.label == expected_ct_label
            assert res.prediction.score == 1.0

            # prediction via stream
            res = m.identify_stream(io.BytesIO(content))
            assert res.path == Path("-")
            assert res.ok
            assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
            assert res.prediction.output.label == expected_ct_label
            assert res.prediction.score == 1.0


def test_magika_module_with_python_and_non_python_content() -> None:
    python_content = (
        b"import flask\nimport requests\n\ndef foo(a):\n    print(f'Test {a}')\n"
    )
    non_python_content = b"clearly not python"

    m = Magika()

    res = m.identify_bytes(python_content)
    assert res.ok
    assert res.prediction.output.label == ContentTypeLabel.PYTHON

    res = m.identify_bytes(non_python_content)
    assert res.ok
    assert res.prediction.output.label == ContentTypeLabel.TXT


def test_magika_module_with_different_prediction_modes() -> None:
    model_dir = utils.get_default_model_dir()
    m = Magika(model_dir=model_dir, prediction_mode=PredictionMode.BEST_GUESS)
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.01) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.40) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.60) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.99) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )

    m = Magika(model_dir=model_dir, prediction_mode=PredictionMode.MEDIUM_CONFIDENCE)
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.01) == (
        ContentTypeLabel.TXT,
        OverwriteReason.LOW_CONFIDENCE,
    )
    assert m._get_output_ct_label_from_dl_result(
        ContentTypeLabel.PYTHON, m._model_config.medium_confidence_threshold - 0.01
    ) == (ContentTypeLabel.TXT, OverwriteReason.LOW_CONFIDENCE)
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.60) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.99) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )

    m = Magika(model_dir=model_dir, prediction_mode=PredictionMode.HIGH_CONFIDENCE)
    high_confidence_threshold = m._model_config.thresholds.get(
        ContentTypeLabel.PYTHON, m._model_config.medium_confidence_threshold
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.01) == (
        ContentTypeLabel.TXT,
        OverwriteReason.LOW_CONFIDENCE,
    )
    assert m._get_output_ct_label_from_dl_result(
        ContentTypeLabel.PYTHON, high_confidence_threshold - 0.01
    ) == (ContentTypeLabel.TXT, OverwriteReason.LOW_CONFIDENCE)
    assert m._get_output_ct_label_from_dl_result(
        ContentTypeLabel.PYTHON, high_confidence_threshold + 0.01
    ) == (ContentTypeLabel.PYTHON, OverwriteReason.NONE)
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.99) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )

    # test that the default is HIGH_CONFIDENCE
    m = Magika(model_dir=model_dir)
    high_confidence_threshold = m._model_config.thresholds.get(
        ContentTypeLabel.PYTHON, m._model_config.medium_confidence_threshold
    )
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.01) == (
        ContentTypeLabel.TXT,
        OverwriteReason.LOW_CONFIDENCE,
    )
    assert m._get_output_ct_label_from_dl_result(
        ContentTypeLabel.PYTHON, high_confidence_threshold - 0.01
    ) == (ContentTypeLabel.TXT, OverwriteReason.LOW_CONFIDENCE)
    assert m._get_output_ct_label_from_dl_result(
        ContentTypeLabel.PYTHON, high_confidence_threshold + 0.01
    ) == (ContentTypeLabel.PYTHON, OverwriteReason.NONE)
    assert m._get_output_ct_label_from_dl_result(ContentTypeLabel.PYTHON, 0.99) == (
        ContentTypeLabel.PYTHON,
        OverwriteReason.NONE,
    )


def test_magika_module_with_directory() -> None:
    m = Magika()

    with tempfile.TemporaryDirectory() as td:
        td_path = Path(td)
        res = m.identify_path(td_path)
        assert res.path == td_path
        assert res.ok
        assert res.prediction.dl.label == ContentTypeLabel.UNDEFINED
        assert res.prediction.output.label == ContentTypeLabel.DIRECTORY
        assert res.prediction.score == 1.0


def test_magika_module_multiple_copies_of_the_same_file() -> None:
    with tempfile.TemporaryDirectory() as td:
        test_path = Path(td) / "test.txt"
        test_path.write_text("test")

        test_paths = [test_path] * 3

        m = Magika()
        results = m.identify_paths(test_paths)
        assert len(results) == len(test_paths)
        for result in results:
            assert result.path == test_path
            assert result.ok
            assert result.prediction.output.label == ContentTypeLabel.TXT


def test_magika_module_with_symlink() -> None:
    with tempfile.TemporaryDirectory() as td:
        test_path = Path(td) / "test.txt"
        test_path.write_text("test")

        symlink_path = Path(td) / "symlink-test.txt"
        symlink_path.symlink_to(test_path)

        m = Magika()
        res = m.identify_path(test_path)
        assert res.path == test_path
        assert res.ok
        assert res.prediction.output.label == ContentTypeLabel.TXT
        res = m.identify_path(symlink_path)
        assert res.path == symlink_path
        assert res.ok
        assert res.prediction.output.label == ContentTypeLabel.TXT

        m = Magika(no_dereference=True)
        res = m.identify_path(test_path)
        assert res.path == test_path
        assert res.ok
        assert res.prediction.output.label == ContentTypeLabel.TXT
        res = m.identify_path(symlink_path)
        assert res.path == symlink_path
        assert res.ok
        assert res.prediction.output.label == ContentTypeLabel.SYMLINK


def test_magika_module_with_non_existing_file() -> None:
    m = Magika()

    with tempfile.TemporaryDirectory() as td:
        non_existing_path = Path(td) / "non_existing.txt"

        res = m.identify_path(non_existing_path)
        assert res.path == non_existing_path
        assert not res.ok
        assert res.status == Status.FILE_NOT_FOUND_ERROR


def test_magika_module_with_permission_error() -> None:
    m = Magika()

    with tempfile.TemporaryDirectory() as td:
        unreadable_test_path = Path(td) / "test.txt"
        unreadable_test_path.write_text("text")

        unreadable_test_path.chmod(0o000)

        res = m.identify_path(unreadable_test_path)
        assert res.path == unreadable_test_path
        assert not res.ok
        assert res.status == Status.PERMISSION_ERROR


@pytest.mark.skip
def test_magika_module_with_really_many_files() -> None:
    test_file_path = utils.get_one_basic_test_file_path()

    m = Magika()

    for n in [10000]:
        test_files_paths = [test_file_path] * n

        results = m.identify_paths(test_files_paths)
        for result in results:
            assert result.ok
            # TODO: add more checks


@pytest.mark.slow
def test_magika_module_with_big_file() -> None:
    def signal_handler(signum: int, frame: Any) -> None:
        raise Exception("Timeout")

    signal.signal(signal.SIGALRM, signal_handler)

    # It should take much less than this, but pytest weird scheduling sometimes
    # creates unexpected slow downs.
    timeout = 2

    m = Magika()

    for sample_size in [1000, 10000, 1_000_000, 1_000_000_000, 10_000_000_000]:
        with tempfile.TemporaryDirectory() as td:
            sample_path = Path(td) / "sample.dat"
            utils.write_random_file_with_size(sample_path, sample_size)
            print(f"Starting running Magika with a timeout of {timeout}")
            signal.alarm(timeout)
            res = m.identify_path(sample_path)
            assert res.ok
            signal.alarm(0)
            print("Done running Magika")


def test_api_call_with_bad_types() -> None:
    m = Magika()

    _ = m.identify_path(Path("/non_existing.txt"))
    _ = m.identify_path("/non_existing.txt")
    with pytest.raises(TypeError):
        _ = m.identify_path(b"/non_existing.txt")  # type: ignore[arg-type]

    _ = m.identify_paths([Path("/non_existing.txt")])
    _ = m.identify_paths(["/non_existing.txt"])
    _ = m.identify_paths([Path("/non_existing.txt"), Path("/not_existing2.txt")])
    _ = m.identify_paths([Path("/non_existing.txt"), "/not_existing2.txt"])
    _ = m.identify_paths(["/non_existing.txt", "/not_existing2.txt"])
    with pytest.raises(TypeError):
        _ = m.identify_paths(Path("/non_existing.txt"))  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        _ = m.identify_paths([b"/non_existing.txt"])  # type: ignore[list-item]
    with pytest.raises(TypeError):
        _ = m.identify_paths([Path("/non_existing.txt"), b"/not_existing2.txt"])  # type: ignore[list-item]

    _ = m.identify_bytes(b"bytes content")
    with pytest.raises(TypeError):
        _ = m.identify_bytes("str content")  # type: ignore[arg-type]

    _ = m.identify_stream(io.BytesIO(b"bytes stream content"))
    with pytest.raises(TypeError):
        _ = m.identify_stream(io.StringIO("str stream content"))  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        _ = m.identify_stream(b"bytes content")  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        _ = m.identify_stream("str content")  # type: ignore[arg-type]


def test_access_magika_result_and_prediction():
    m = Magika()

    res = m.identify_bytes(b"text")
    assert isinstance(res, MagikaResult)
    assert isinstance(res.path, Path)
    assert isinstance(res.ok, bool)
    assert isinstance(res.status, Status)
    assert isinstance(res.prediction, MagikaPrediction)
    assert isinstance(res.prediction.dl, ContentTypeInfo)
    assert isinstance(res.prediction.output, ContentTypeInfo)
    assert isinstance(res.prediction.score, float)
    # test access to forwarded properties
    assert isinstance(res.dl, ContentTypeInfo)
    assert isinstance(res.output, ContentTypeInfo)
    assert isinstance(res.score, float)
    # test access to non-existing properties
    with pytest.raises(AttributeError):
        _ = res.foo  # type: ignore[attr-defined]
    with pytest.raises(AttributeError):
        _ = res.prediction.foo  # type: ignore[attr-defined]

    res = m.identify_path(Path("/non_existing.txt"))
    assert isinstance(res, MagikaResult)
    assert isinstance(res.path, Path)
    assert isinstance(res.ok, bool)
    assert isinstance(res.status, Status)
    with pytest.raises(ValueError):
        _ = res.prediction
    with pytest.raises(ValueError):
        _ = res.prediction.dl
    with pytest.raises(ValueError):
        _ = res.prediction.output
    with pytest.raises(ValueError):
        _ = res.prediction.score
    with pytest.raises(ValueError):
        _ = res.dl
    with pytest.raises(ValueError):
        _ = res.output
    with pytest.raises(ValueError):
        _ = res.score
    with pytest.raises(AttributeError):
        _ = res.foo  # type: ignore[attr-defined]
    with pytest.raises(ValueError):
        _ = res.prediction.foo  # type: ignore[attr-defined]


def test_access_backward_compatibility_layer() -> None:
    m = Magika()

    res = m.identify_bytes(b"text")
    assert isinstance(res, MagikaResult)
    assert isinstance(res.path, Path)
    assert isinstance(res.ok, bool)
    assert isinstance(res.status, Status)
    assert isinstance(res.prediction, MagikaPrediction)
    assert isinstance(res.prediction.dl, ContentTypeInfo)
    assert isinstance(res.prediction.output, ContentTypeInfo)
    assert isinstance(res.prediction.score, float)

    with pytest.warns(DeprecationWarning):
        assert res.dl.ct_label == res.prediction.dl.label
    with pytest.warns(DeprecationWarning):
        assert res.output.ct_label == res.prediction.output.label

    with pytest.raises(AttributeError):
        _ = res.dl.score
    with pytest.raises(AttributeError):
        _ = res.output.score

    with pytest.warns(DeprecationWarning):
        assert res.dl.magic == res.prediction.dl.description
    with pytest.warns(DeprecationWarning):
        assert res.output.magic == res.prediction.output.description


def test_get_model_and_output_content_types() -> None:
    m = Magika()
    output_content_types = m.get_output_content_types()
    output_content_types_set = set(output_content_types)
    model_content_types = m.get_model_content_types()
    model_content_types_set = set(model_content_types)

    assert isinstance(output_content_types, List)
    assert len(output_content_types) > 0
    assert isinstance(model_content_types, List)
    assert len(model_content_types) > 0

    for ct in output_content_types:
        assert isinstance(ct, ContentTypeLabel)

    # Check for no duplicates
    assert len(output_content_types) == len(output_content_types_set)

    # Check basic properties about special ContentTypeLabel entries
    special_output_content_types = {
        ContentTypeLabel.DIRECTORY,
        ContentTypeLabel.EMPTY,
        ContentTypeLabel.SYMLINK,
        ContentTypeLabel.TXT,
        ContentTypeLabel.UNKNOWN,
    }
    special_model_content_types = {ContentTypeLabel.UNDEFINED}
    assert special_output_content_types.issubset(output_content_types_set)
    assert not special_model_content_types.issubset(output_content_types_set)
    assert special_model_content_types.issubset(model_content_types_set)
    assert not special_output_content_types.issubset(model_content_types_set)

    # Spot check for popular content types
    assert {
        ContentTypeLabel.ELF,
        ContentTypeLabel.PDF,
    }.issubset(output_content_types_set)
    assert {
        ContentTypeLabel.ELF,
        ContentTypeLabel.PDF,
    }.issubset(model_content_types_set)


def get_expected_content_type_label_from_test_file_path(
    test_path: Path,
) -> ContentTypeLabel:
    return ContentTypeLabel(test_path.parent.name)


def check_result_vs_expected_result(
    file_path: Path, result: MagikaResult, expected_result_path: Optional[Path] = None
) -> None:
    if expected_result_path is None:
        expected_result_path = file_path
    assert result.path == expected_result_path
    assert result.ok
    expected_ct_label = get_expected_content_type_label_from_test_file_path(file_path)
    assert result.prediction.output.label == expected_ct_label


def check_results_vs_expected_results(
    files_paths: List[Path], results: List[MagikaResult]
) -> None:
    for file_path, result in zip(files_paths, results):
        check_result_vs_expected_result(file_path, result)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Taken from https://en.wikipedia.org/wiki/ANSI_escape_code

BLACK = "\033[0;30m"
RED = "\033[0;31m"
GREEN = "\033[0;32m"
YELLOW = "\033[0;33m"
BLUE = "\033[0;34m"
PURPLE = "\033[0;35m"
CYAN = "\033[0;36m"
LIGHT_GRAY = "\033[0;37m"

DARK_GRAY = "\033[1;30m"
LIGHT_RED = "\033[1;31m"
LIGHT_GREEN = "\033[1;32m"
LIGHT_YELLOW = "\033[1;33m"
LIGHT_BLUE = "\033[1;34m"
LIGHT_PURPLE = "\033[1;35m"
LIGHT_CYAN = "\033[1;36m"
WHITE = "\033[1;37m"

RESET = "\033[0;39m"



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from pathlib import Path


def test_python_magika_client() -> None:
    python_root_dir = Path(__file__).parent.parent
    python_magika_client_path = (
        python_root_dir / "src" / "magika" / "cli" / "magika_client.py"
    ).resolve()

    # quick test to check there are no obvious problems
    cmd = [str(python_magika_client_path), "--help"]
    subprocess.run(cmd, capture_output=True, check=True)

    # quick test to check there are no crashes
    cmd = [str(python_magika_client_path), str(python_magika_client_path)]
    subprocess.run(cmd, capture_output=True, check=True)



#!/usr/bin/env python
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


def main() -> None:
    message = """
WARNING: you have attempted to run `$ magika` (the Rust client), but this is not
available in the python package you installed, likely because magika pipeline
does not currently build binary wheels compatible with your platform settings.

If you think this is a problem worth solving, please open an issue at
https://github.com/google/magika.

In the meantime, you can use the old python magika client with `$ magika-python-client`.
"""

    print(message.strip())
    sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2023-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script should only rely on dependencies installed with `pip install
magika`; this script is used as part of "build & test package" github action,
and the dev dependencies are not available.
"""

import sys
from pathlib import Path

import click

from magika import ContentTypeLabel, Magika, PredictionMode


@click.command()
def main() -> None:
    m = Magika(prediction_mode=PredictionMode.HIGH_CONFIDENCE)

    print(f"Magika instance details: {m}")

    res = m.identify_bytes(b"text")
    assert res.dl.label == ContentTypeLabel.UNDEFINED
    assert res.output.label == ContentTypeLabel.TXT
    assert res.score == 1.0

    res = m.identify_bytes(b"\xff\xff\xff")
    assert res.dl.label == ContentTypeLabel.UNDEFINED
    assert res.output.label == ContentTypeLabel.UNKNOWN
    assert res.score == 1.0

    basic_tests_dir = (
        Path(__file__).parent.parent.parent / "tests_data" / "basic"
    ).resolve()

    files_paths = sorted(filter(lambda p: p.is_file(), basic_tests_dir.rglob("*")))

    with_error = False
    for file_path in files_paths:
        res = m.identify_path(file_path)
        output_label = res.output.label
        expected_label = file_path.parent.name
        if expected_label != output_label:
            with_error = True
            print(
                f"ERROR: Misprediction for {file_path}: expected_label={expected_label}, output_label={output_label}"
            )

    if with_error:
        print("ERROR: There was at least one misprediction")
        sys.exit(1)

    print("All examples were predicted correctly")


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import sys
from typing import Optional, TextIO

from magika import colors

_logger: Optional[SimpleLogger] = None


class SimpleLogger:
    """
    We implement a simple logger to not rely on additional python packages,
    e.g., rich. This is written in way that, by default, log messages (e.g.,
    debug/info/...) are sent to stderr.
    """

    def __init__(self, use_colors: bool = False):
        self.level = logging.WARNING
        self.use_colors = use_colors

    def setLevel(self, level: int) -> None:
        self.level = level

    def raw_print_to_stdout(self, msg: str) -> None:
        self.raw_print(msg, file=sys.stdout)

    def raw_print(
        self, msg: str, file: Optional[TextIO] = None, flush: bool = True
    ) -> None:
        if file is None:
            # We avoid using a default value for the `file` argument because we
            # need to get the reference to the "current" stderr; if we used a
            # default argument, we would just store the "current at
            # instantiation time" stderr, which may not be the current one.
            # This, in turn, could create problems for testing.
            file = sys.stderr
        print(msg, file=file, flush=flush)

    def debug(self, msg: str) -> None:
        if logging.DEBUG >= self.level:
            if self.use_colors:
                self.raw_print(f"{colors.GREEN}DEBUG: {msg}{colors.RESET}")
            else:
                self.raw_print(f"DEBUG: {msg}")

    def info(self, msg: str) -> None:
        if logging.INFO >= self.level:
            self.raw_print(f"INFO: {msg}")

    def warning(self, msg: str) -> None:
        if logging.WARNING >= self.level:
            if self.use_colors:
                self.raw_print(f"{colors.YELLOW}WARNING: {msg}{colors.RESET}")
            else:
                self.raw_print(f"WARNING: {msg}")

    def error(self, msg: str) -> None:
        if logging.ERROR >= self.level:
            if self.use_colors:
                self.raw_print(f"{colors.RED}ERROR: {msg}{colors.RESET}")
            else:
                self.raw_print(f"ERROR: {msg}")


def get_logger(use_colors: bool = False) -> SimpleLogger:
    global _logger

    if _logger is None:
        _logger = SimpleLogger(use_colors=use_colors)

    return _logger



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path

import tomli
import tomli_w


def main() -> None:
    pyproject_toml_path = Path(__file__).parent.parent / "pyproject.toml"

    pyproject_content = tomli.loads(pyproject_toml_path.read_text())

    # Remove entry about maturin, we don't need it
    _ = pyproject_content["tool"].pop("maturin")

    # Tell uv we want to use the hatchling build system
    pyproject_content["build-system"] = {
        "requires": ["hatchling"],
        "build-backend": "hatchling.build",
    }

    # Make the python's magika client available as a script
    pyproject_content["project"]["scripts"] = {
        "magika-python-client": "magika.cli.magika_client:main",
        "magika": "magika.cli.magika_rust_client_not_found_warning:main",
    }

    pyproject_toml_path.write_text(tomli_w.dumps(pyproject_content))


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import Dict, List

from magika.types.content_type_label import ContentTypeLabel


@dataclass(frozen=True)
class ModelFeatures:
    beg: List[int]
    mid: List[int]
    end: List[int]
    # for ISO
    offset_0x8000_0x8007: List[int]
    offset_0x8800_0x8807: List[int]
    offset_0x9000_0x9007: List[int]
    # for UDF
    offset_0x9800_0x9807: List[int]


@dataclass(frozen=True)
class ModelOutput:
    ct_label: ContentTypeLabel
    score: float


@dataclass(frozen=True)
class ModelConfig:
    beg_size: int
    mid_size: int
    end_size: int
    use_inputs_at_offsets: bool
    medium_confidence_threshold: float
    min_file_size_for_dl: int
    padding_token: int
    block_size: int
    target_labels_space: List[ContentTypeLabel]
    thresholds: Dict[ContentTypeLabel, float]
    overwrite_map: Dict[ContentTypeLabel, ContentTypeLabel]



#!/usr/bin/env python
# Copyright 2023-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script assumes that the `magika` python package has been already installed,
and that the `magika` client is available in the PATH. The script tests that the
`magika` client appears functional.

This script should only rely on dependencies installed with `pip install
magika`; this script is used as part of "build & test package" github action,
and the dev dependencies are not available.
"""

import subprocess
import sys
from pathlib import Path
from typing import Optional

import click


@click.command()
@click.option(
    "--client-path",
    type=click.Path(exists=False, resolve_path=False),
)
def main(client_path: Optional[Path]) -> None:
    """Tests the Rust or Python Magika client. By default, it runs "magika"
    (expected in PATH). Use --client-path to specify a different client
    executable. Note that "client_path" may not point to a valid file, but still
    be a valid target to run as it may be found in the PATH.
    """

    basic_tests_dir = (
        Path(__file__).resolve().parent.parent.parent / "tests_data" / "basic"
    )
    assert basic_tests_dir.is_dir()

    if client_path is None:
        client_path = Path("magika")
    print(f'Testing client: "{client_path}"')

    p = subprocess.run([str(client_path), "--version"], capture_output=True, text=True)
    print(f'Output of "magika --version": {p.stdout.strip()}')

    p = subprocess.run(
        [str(client_path), "-r", "--label", "--no-colors", str(basic_tests_dir)],
        capture_output=True,
        text=True,
    )

    if p.returncode != 0:
        print("ERROR: magika CLI exited with non-zero status.")
        print(f"stdout:\n{p.stdout}\n" + "-" * 40)
        print(f"stderr:\n{p.stderr}\n" + "-" * 40)
        sys.exit(1)

    if p.stderr != "":
        print(f"WARNING: p.stderr not empty: {p.stderr}")

    with_error = False
    lines = p.stdout.split("\n")
    for line in lines:
        line = line.strip()
        if line == "":
            continue
        file_path_str, file_output_str = line.split(": ", 1)
        file_path = Path(file_path_str)
        output_label = file_output_str.strip().split(" ", 1)[0]
        expected_label = file_path.parent.name
        if expected_label != output_label:
            with_error = True
            print(
                f"ERROR: Misprediction for {file_path}: expected_label={expected_label}, output_label={output_label}"
            )

    if with_error:
        print("ERROR: There was at least one misprediction")
        sys.exit(1)

    print("All examples were predicted correctly")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import importlib.metadata
import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import click

from magika import Magika, MagikaError, PredictionMode, colors
from magika.logger import get_logger
from magika.types import ContentTypeLabel, MagikaResult
from magika.types.overwrite_reason import OverwriteReason

VERSION = importlib.metadata.version("magika")

CONTACT_EMAIL = "magika-dev@google.com"

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

HELP_EPILOG = f"""
Magika version: "{VERSION}"\f
Default model: "{Magika._get_default_model_name()}"

Send any feedback to {CONTACT_EMAIL} or via GitHub issues.
"""


@click.command(
    context_settings=CONTEXT_SETTINGS,
    epilog=HELP_EPILOG,
)
@click.argument(
    "file",
    type=click.Path(exists=False, readable=False, path_type=Path),
    required=False,
    nargs=-1,
)
@click.option(
    "-r",
    "--recursive",
    is_flag=True,
    help='When passing this option, magika scans every file within directories, instead of outputting "directory"',
)
@click.option("--json", "json_output", is_flag=True, help="Output in JSON format.")
@click.option("--jsonl", "jsonl_output", is_flag=True, help="Output in JSONL format.")
@click.option(
    "-i",
    "--mime-type",
    "mime_output",
    is_flag=True,
    help="Output the MIME type instead of a verbose content type description.",
)
@click.option(
    "-l",
    "--label",
    "label_output",
    is_flag=True,
    help="Output a simple label instead of a verbose content type description. Use --list-output-content-types for the list of supported output.",
)
@click.option(
    "-c",
    "--compatibility-mode",
    "magic_compatibility_mode",
    is_flag=True,
    help="Compatibility mode: output is as close as possible to `file` and colors are disabled.",
)
@click.option(
    "-s",
    "--output-score",
    is_flag=True,
    help="Output the prediction's score in addition to the content type.",
)
@click.option(
    "-m",
    "--prediction-mode",
    "prediction_mode_str",
    type=click.Choice(PredictionMode.get_valid_prediction_modes(), case_sensitive=True),
    default=PredictionMode.HIGH_CONFIDENCE,
)
@click.option(
    "--batch-size", default=32, help="How many files to process in one batch."
)
@click.option(
    "--no-dereference",
    is_flag=True,
    help="This option causes symlinks not to be followed. By default, symlinks are dereferenced.",
)
@click.option(
    "--colors/--no-colors",
    "with_colors",
    is_flag=True,
    default=True,
    help="Enable/disable use of colors.",
)
@click.option("-v", "--verbose", is_flag=True, help="Enable more verbose output.")
@click.option("-vv", "--debug", is_flag=True, help="Enable debug logging.")
@click.option(
    "--version", "output_version", is_flag=True, help="Print the version and exit."
)
@click.option(
    "--model-dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path
    ),
    help="Use a custom model.",
)
def main(
    file: List[Path],
    recursive: bool,
    json_output: bool,
    jsonl_output: bool,
    mime_output: bool,
    label_output: bool,
    magic_compatibility_mode: bool,
    output_score: bool,
    prediction_mode_str: str,
    batch_size: int,
    no_dereference: bool,
    with_colors: bool,
    verbose: bool,
    debug: bool,
    output_version: bool,
    model_dir: Optional[Path],
) -> None:
    """
    Magika - Determine type of FILEs with deep-learning.
    """

    # click uses the name of the variable to determine how it will show up in
    # the --help. Since we don't like to see "file_paths" in the help, we name
    # the argument "file" (which is ugly) and we re-assign it as soon as we can.
    files_paths = file

    if magic_compatibility_mode:
        # In compatibility mode we disable colors.
        with_colors = False

    _l = get_logger(use_colors=with_colors)

    if verbose:
        _l.setLevel(logging.INFO)
    if debug:
        _l.setLevel(logging.DEBUG)

    if output_version:
        _l.raw_print_to_stdout("Magika python client")
        _l.raw_print_to_stdout(f"Magika version: {VERSION}")
        _l.raw_print_to_stdout(f"Default model: {Magika._get_default_model_name()}")
        sys.exit(0)

    if len(files_paths) == 0:
        _l.error("You need to pass at least one path, or - to read from stdin.")
        sys.exit(1)

    read_from_stdin = False
    for p in files_paths:
        if str(p) == "-":
            read_from_stdin = True
        elif not p.exists():
            _l.error(f'File or directory "{str(p)}" does not exist.')
            sys.exit(1)
    if read_from_stdin:
        if len(files_paths) > 1:
            _l.error('If you pass "-", you cannot pass anything else.')
            sys.exit(1)
        if recursive:
            _l.error('If you pass "-", recursive scan is not meaningful.')
            sys.exit(1)

    if batch_size <= 0 or batch_size > 512:
        _l.error("Batch size needs to be greater than 0 and less or equal than 512.")
        sys.exit(1)

    if json_output and jsonl_output:
        _l.error("You should use either --json or --jsonl, not both.")
        sys.exit(1)

    if int(mime_output) + int(label_output) + int(magic_compatibility_mode) > 1:
        _l.error("You should use only one of --mime, --label, --compatibility-mode.")
        sys.exit(1)

    if recursive:
        # recursively enumerate files within directories
        expanded_paths = []
        for p in files_paths:
            if p.exists():
                if p.is_file():
                    expanded_paths.append(p)
                elif p.is_dir():
                    expanded_paths.extend(sorted(p.rglob("*")))
            elif str(p) == "-":
                # this is "read from stdin", that's OK
                pass
            else:
                _l.error(f'File or directory "{str(p)}" does not exist.')
                sys.exit(1)
        # the resulting list may still include some directories; thus, we filter them out.
        files_paths: List[Path] = list(filter(lambda x: not x.is_dir(), expanded_paths))  # type: ignore[no-redef]

    _l.info(f"Considering {len(files_paths)} files")
    _l.debug(f"Files: {files_paths}")

    # Select an alternative model checking: 1) CLI option, 2) env variable.
    # If none of these is set, model_dir is left to None, and the Magika module
    # will use the default model.
    if model_dir is None:
        model_dir_str = os.environ.get("MAGIKA_MODEL_DIR")
        if model_dir_str is not None and model_dir_str.strip() != "":
            model_dir = Path(model_dir_str)

    try:
        magika = Magika(
            model_dir=model_dir,
            prediction_mode=PredictionMode(prediction_mode_str),
            no_dereference=no_dereference,
            verbose=verbose,
            debug=debug,
            use_colors=with_colors,
        )
    except MagikaError as mr:
        _l.error(str(mr))
        sys.exit(1)

    start_color = ""
    end_color = ""

    color_by_group = {
        "document": colors.LIGHT_PURPLE,
        "executable": colors.LIGHT_GREEN,
        "archive": colors.LIGHT_RED,
        "audio": colors.YELLOW,
        "image": colors.YELLOW,
        "video": colors.YELLOW,
        "code": colors.LIGHT_BLUE,
    }

    # updated only when we need to output in JSON format
    all_predictions: List[Tuple[Path, MagikaResult]] = []

    batches_num = len(files_paths) // batch_size
    if len(files_paths) % batch_size != 0:
        batches_num += 1
    for batch_idx in range(batches_num):
        batch_files_paths = files_paths[
            batch_idx * batch_size : (batch_idx + 1) * batch_size
        ]

        if should_read_from_stdin(files_paths):
            batch_predictions = [get_magika_result_from_stdin(magika)]
        else:
            batch_predictions = magika.identify_paths(batch_files_paths)

        if json_output:
            # we do not stream the output for JSON output
            all_predictions.extend(zip(batch_files_paths, batch_predictions))
        elif jsonl_output:
            for file_path, result in zip(batch_files_paths, batch_predictions):
                _l.raw_print_to_stdout(json.dumps(result_to_dict(result)))
        else:
            for file_path, result in zip(batch_files_paths, batch_predictions):
                if result.ok:
                    if mime_output:
                        # If the user requested the MIME type, we use the mime type
                        # regardless of the compatibility mode.
                        output = result.prediction.output.mime_type
                    elif label_output:
                        output = str(result.prediction.output.label)
                    else:  # human-readable description
                        output = f"{result.prediction.output.description} ({result.prediction.output.group})"

                        if (
                            result.prediction.dl.label != ContentTypeLabel.UNDEFINED
                            and result.prediction.dl.label
                            != result.prediction.output.label
                            and result.prediction.overwrite_reason
                            == OverwriteReason.NONE
                        ):
                            # It seems that we had a low-confidence prediction
                            # from the model. Let's warn the user about our best
                            # bet.
                            output += (
                                " [Low-confidence model best-guess: "
                                f"{result.prediction.dl.description} ({result.prediction.dl.group}), "
                                f"score={result.prediction.score}]"
                            )

                    if with_colors:
                        start_color = color_by_group.get(
                            result.prediction.output.group, colors.WHITE
                        )
                        end_color = colors.RESET
                else:
                    output = result.status
                    start_color = ""
                    end_color = ""

                if output_score and result.ok:
                    score = int(result.prediction.score * 100)
                    _l.raw_print_to_stdout(
                        f"{start_color}{file_path}: {output} {score}%{end_color}"
                    )
                else:
                    _l.raw_print_to_stdout(
                        f"{start_color}{file_path}: {output}{end_color}"
                    )

    if json_output:
        _l.raw_print_to_stdout(
            json.dumps(
                [result_to_dict(result) for (_, result) in all_predictions],
                indent=4,
            )
        )


def should_read_from_stdin(files_paths: List[Path]) -> bool:
    return len(files_paths) == 1 and str(files_paths[0]) == "-"


def get_magika_result_from_stdin(magika: Magika) -> MagikaResult:
    content = sys.stdin.buffer.read()
    result = magika.identify_bytes(content)
    return result


def result_to_dict(result: MagikaResult) -> dict:
    out: Dict[str, Any] = {"path": str(result.path), "status": result.status}
    if result.ok:
        out["prediction"] = dataclasses.asdict(result.prediction)
    return out


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import json
import logging
import os
import time
from pathlib import Path
from typing import BinaryIO, Dict, List, Optional, Sequence, Set, Tuple, Union

import numpy as np
import numpy.typing as npt
import onnxruntime as rt

from magika.logger import get_logger
from magika.seekable import Buffer, File, Seekable, Stream
from magika.types import (
    ContentTypeInfo,
    ContentTypeLabel,
    MagikaError,
    MagikaPrediction,
    MagikaResult,
    ModelConfig,
    ModelFeatures,
    ModelOutput,
    OverwriteReason,
    PredictionMode,
    Status,
)

_DEFAULT_MODEL_NAME = "standard_v3_2"


class Magika:
    def __init__(
        self,
        model_dir: Optional[Path] = None,
        prediction_mode: PredictionMode = PredictionMode.HIGH_CONFIDENCE,
        no_dereference: bool = False,
        verbose: bool = False,
        debug: bool = False,
        use_colors: bool = False,
    ) -> None:
        self._log = get_logger(use_colors=use_colors)

        if verbose:
            self._log.setLevel(logging.INFO)

        if debug:
            self._log.setLevel(logging.DEBUG)

        if model_dir is not None:
            self._model_dir = model_dir
        else:
            # use default model
            self._model_dir = (
                Path(__file__).parent / "models" / self._get_default_model_name()
            )

        self._model_path = self._model_dir / "model.onnx"
        self._model_config_path = self._model_dir / "config.min.json"

        if not self._model_dir.is_dir():
            raise MagikaError(f"model dir not found at {str(self._model_dir)}")
        if not self._model_path.is_file():
            raise MagikaError(f"model not found at {str(self._model_path)}")
        if not self._model_config_path.is_file():
            raise MagikaError(
                f"model config not found at {str(self._model_config_path)}"
            )

        self._model_config: ModelConfig = Magika._load_model_config(
            self._model_config_path
        )

        self._target_labels_space_np = np.array(
            list(map(str, self._model_config.target_labels_space))
        )

        self._prediction_mode = prediction_mode

        self._no_dereference = no_dereference

        content_types_kb_path = (
            Path(__file__).parent / "config" / "content_types_kb.min.json"
        )
        self._cts_infos = Magika._load_content_types_kb(content_types_kb_path)

        self._onnx_session = self._init_onnx_session()

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f'Magika(module_version="{self.get_module_version()}", model_name="{self.get_model_name()}")'

    def get_module_version(self) -> str:
        return str(__import__(self.__module__).__version__)

    def get_model_name(self) -> str:
        return self._model_dir.name

    def identify_path(self, path: Union[str, os.PathLike]) -> MagikaResult:
        """Identify the content type of a file given its path."""

        if isinstance(path, str) or isinstance(path, os.PathLike):
            path = Path(path)
        else:
            raise TypeError(
                f"Path '{path}' is invalid: input path should be of type `Union[str, os.PathLike]`"
            )

        return self._get_result_from_path(path)

    def identify_paths(
        self, paths: Sequence[Union[str, os.PathLike]]
    ) -> List[MagikaResult]:
        """Identify the content type of a list of files given their paths."""

        if not isinstance(paths, Sequence):
            raise TypeError("Input paths should be of type Sequence[Path]")

        paths_ = []
        for path in paths:
            if isinstance(path, str) or isinstance(path, os.PathLike):
                paths_.append(Path(path))
            else:
                raise TypeError(
                    f"Input '{path}' is invalid: input path should be of type `Union[str, os.PathLike]`"
                )

        return self._get_results_from_paths(paths_)

    def identify_bytes(self, content: bytes) -> MagikaResult:
        """Identify the content type of raw bytes."""

        if not isinstance(content, bytes):
            raise TypeError(
                f"Input content should be of type 'bytes', not {type(content)}."
            )
        return self._get_result_from_bytes(content)

    def identify_stream(self, stream: BinaryIO) -> MagikaResult:
        """Identify the content type of a BinaryIO stream. Note that this method will
        seek() around the stream."""

        if not isinstance(stream, io.IOBase) or not stream.readable():  # type: ignore[unreachable]
            raise TypeError("Input stream must be a readable BinaryIO object.")

        # Explicitly test for the most common error so that we can return an
        # helpful error message.
        if isinstance(stream, io.TextIOBase):  # type: ignore[unreachable]
            raise TypeError(
                "Input stream must be opened in bytes mode, not in text mode."
            )

        if not isinstance(stream, io.BufferedIOBase):
            raise TypeError("Input stream must be a readable BinaryIO object.")

        if (
            not hasattr(stream, "seek")
            or not hasattr(stream, "read")
            or not hasattr(stream, "tell")
        ):
            raise TypeError("Input stream must have seek, read, and tell methods.")

        return self._get_result_from_stream(stream)

    def get_output_content_types(self) -> List[ContentTypeLabel]:
        """This method returns the list of all possible output content types of
        the module. I.e., all possible values for
        `MagikaResult.prediction.output.label`.  This considers the list of
        possible outputs from the model itself, but also keeps into account
        additional configuration such as `override_map` and special content
        types such as `ContentTypeLabel.EMPTY` or `ContentTypeLabel.SYMLINK`.
        """

        target_labels_space = self._model_config.target_labels_space
        overwrite_map = self._model_config.overwrite_map

        output_content_types: Set[ContentTypeLabel] = {
            ContentTypeLabel.DIRECTORY,
            ContentTypeLabel.EMPTY,
            ContentTypeLabel.SYMLINK,
            ContentTypeLabel.TXT,
            ContentTypeLabel.UNKNOWN,
        }
        for ct in target_labels_space:
            # Check if we would overwrite this target label; if not, use the
            # target label itself.
            output_ct = overwrite_map.get(ct, ct)
            output_content_types.add(output_ct)

        return sorted(output_content_types)

    def get_model_content_types(self) -> List[ContentTypeLabel]:
        """This method returns the list of all possible output of the underlying
        model. I.e., all possible values for `MagikaResult.prediction.dl.label`.
        Note that, in general, the list of "model outputs" is different than the
        "tool outputs" as in some cases the model is not even used, or the
        model's output is overwritten due to a low-confidence score, or other
        reasons.  This API is useful mostly for debugging purposes; the vast
        majority of client should use `get_output_content_types()`.
        """

        model_content_types: Set[ContentTypeLabel] = {
            ContentTypeLabel.UNDEFINED,
        }
        model_content_types.update(self._model_config.target_labels_space)
        return sorted(model_content_types)

    @staticmethod
    def _get_default_model_name() -> str:
        """This returns the default model name.

        We make this method static so that it can be used by external
        clients/tests without the need to instantiate a Magika object.
        """

        return _DEFAULT_MODEL_NAME

    @staticmethod
    def _load_content_types_kb(
        content_types_kb_json_path: Path,
    ) -> Dict[ContentTypeLabel, ContentTypeInfo]:
        TXT_MIME_TYPE = "text/plain"
        UNKNOWN_MIME_TYPE = "application/octet-stream"
        UNKNOWN_GROUP = "unknown"

        out = {}
        for ct_name, ct_info in json.loads(
            content_types_kb_json_path.read_text()
        ).items():
            is_text = ct_info["is_text"]
            if is_text:
                default_mime_type = TXT_MIME_TYPE
            else:
                default_mime_type = UNKNOWN_MIME_TYPE
            mime_type = (
                default_mime_type
                if ct_info["mime_type"] is None
                else ct_info["mime_type"]
            )
            group = UNKNOWN_GROUP if ct_info["group"] is None else ct_info["group"]
            description = (
                ct_name if ct_info["description"] is None else ct_info["description"]
            )
            extensions = ct_info["extensions"]
            out[ContentTypeLabel(ct_name)] = ContentTypeInfo(
                label=ContentTypeLabel(ct_name),
                mime_type=mime_type,
                group=group,
                description=description,
                extensions=extensions,
                is_text=is_text,
            )
        return out

    @staticmethod
    def _load_model_config(model_config_path: Path) -> ModelConfig:
        config = json.loads(model_config_path.read_text())

        return ModelConfig(
            beg_size=config["beg_size"],
            mid_size=config["mid_size"],
            end_size=config["end_size"],
            use_inputs_at_offsets=config["use_inputs_at_offsets"],
            medium_confidence_threshold=config["medium_confidence_threshold"],
            min_file_size_for_dl=config["min_file_size_for_dl"],
            padding_token=config["padding_token"],
            block_size=config["block_size"],
            target_labels_space=[
                ContentTypeLabel(ct_str) for ct_str in config["target_labels_space"]
            ],
            thresholds={
                ContentTypeLabel(k): v for k, v in config["thresholds"].items()
            },
            overwrite_map={
                ContentTypeLabel(k): ContentTypeLabel(v)
                for k, v in config["overwrite_map"].items()
            },
        )

    def _init_onnx_session(self) -> rt.InferenceSession:
        start_time = time.time()
        rt.disable_telemetry_events()

        onnx_session = rt.InferenceSession(
            self._model_path,
            providers=["CPUExecutionProvider"],
        )
        elapsed_time = 1000 * (time.time() - start_time)
        self._log.debug(
            f'ONNX DL model "{self._model_path}" loaded in {elapsed_time:.03f} ms'
        )
        return onnx_session

    def _get_ct_info(self, content_type: ContentTypeLabel) -> ContentTypeInfo:
        return self._cts_infos[content_type]

    def _get_results_from_paths(self, paths: List[Path]) -> List[MagikaResult]:
        """Given a list of paths, returns a list of MagikaResult objects, which
        contain relevant information, such as: file path, the output of the DL
        model, the confidence score, the output of the tool, and associated
        metadata. The order of the predictions matches the order of the input
        paths."""

        # We do a first pass on all files: we collect features for the files
        # that need to be analyzed with the DL model, and we already determine
        # the output for the remaining ones.

        # We use a "str" instead of Path because it makes it easier later on to
        # serialize.
        all_outputs: Dict[str, MagikaResult] = {}  # {path: <output>, ...}

        # We use a list and not the dict because that's what we need later on
        # for inference.
        all_features: List[Tuple[Path, ModelFeatures]] = []

        self._log.debug(
            f"Processing input files and extracting features for {len(paths)} samples"
        )
        start_time = time.time()
        for path in paths:
            output, features = self._get_result_or_features_from_path(path)
            if output is not None:
                all_outputs[str(path)] = output
            else:
                assert features is not None
                all_features.append((path, features))
        elapsed_time = 1000 * (time.time() - start_time)
        self._log.debug(f"First pass and features extracted in {elapsed_time:.03f} ms")

        # Get the outputs via DL for the files that need it.
        for path_str, result in self._get_results_from_features(all_features).items():
            all_outputs[path_str] = result

        # Finally, we collect the predictions in a final list, sorted by the
        # initial paths list (and not by insertion order).
        sorted_outputs = []
        for path in paths:
            sorted_outputs.append(all_outputs[str(path)])
        return sorted_outputs

    def _get_result_from_path(self, path: Path) -> MagikaResult:
        return self._get_results_from_paths([path])[0]

    def _get_result_from_bytes(self, content: bytes) -> MagikaResult:
        result, features = self._get_result_or_features_from_bytes(content)
        if result is not None:
            return result
        assert features is not None
        return self._get_result_from_features(features)

    def _get_result_from_stream(self, stream: BinaryIO) -> MagikaResult:
        result, features = self._get_result_or_features_from_stream(stream)
        if result is not None:
            return result
        assert features is not None
        return self._get_result_from_features(features)

    @staticmethod
    def _extract_features_from_path(
        file_path: Path,
        beg_size: int,
        mid_size: int,
        end_size: int,
        padding_token: int,
        block_size: int,
        use_inputs_at_offsets: bool,
    ) -> ModelFeatures:
        # TODO: reimplement this using a context manager
        file = File(file_path)
        mf = Magika._extract_features_from_seekable(
            file,
            beg_size,
            mid_size,
            end_size,
            padding_token,
            block_size,
            use_inputs_at_offsets,
        )
        file.close()
        return mf

    @staticmethod
    def _extract_features_from_bytes(
        content: bytes,
        beg_size: int,
        mid_size: int,
        end_size: int,
        padding_token: int,
        block_size: int,
        use_inputs_at_offsets: bool,
    ) -> ModelFeatures:
        return Magika._extract_features_from_seekable(
            Buffer(content),
            beg_size,
            mid_size,
            end_size,
            padding_token,
            block_size,
            use_inputs_at_offsets,
        )

    @staticmethod
    def _extract_features_from_stream(
        stream: BinaryIO,
        beg_size: int,
        mid_size: int,
        end_size: int,
        padding_token: int,
        block_size: int,
        use_inputs_at_offsets: bool,
    ) -> ModelFeatures:
        return Magika._extract_features_from_seekable(
            Stream(stream),
            beg_size,
            mid_size,
            end_size,
            padding_token,
            block_size,
            use_inputs_at_offsets,
        )

    @staticmethod
    def _extract_features_from_seekable(
        seekable: Seekable,
        beg_size: int,
        mid_size: int,
        end_size: int,
        padding_token: int,
        block_size: int,
        use_inputs_at_offsets: bool,
    ) -> ModelFeatures:
        """This implement v2 of the features extraction v2 from a seekable,
        which is an abstraction about anything that can be "read_at" a specific
        offset, such as a file or buffer. This is implemented so that we do not
        need to load the entire file in memory, or scan the entire buffer.

        High-level overview on what we do:
        - We extract blocks of bytes from the beginning ("beg"), the middle
        ("mid"), and at the end ("end").
        - We then truncate or add padding to these blocks, depending on whether
        we have too many or too few.

        Blocks extraction and padding:
        - beg: we read the first block_size bytes, we lstrip() it, and we use
        this as the basis to extract beg_size integers. If we have too many
        bytes, we only consider the first beg_size ones. If we do not have
        enough, we add padding as suffix (up to beg_size integers).
        - mid: we determine "where the middle is" by using the entire content's
        size (before stripping the whitespace-like characters), and we take the
        mid_size bytes in the middle. If we do not have enough bytes, we add
        padding to the left and to the right. In case we need to add an odd
        number of padding integers, we add an extra one to the right.
        - end: same as "beg", but we read the last block_size bytes, we rstrip()
        (instead of lstrip()), and, if needed, we add padding as a prefix (and
        not as a suffix like we do with "beg").
        """

        assert beg_size < block_size
        assert mid_size < block_size
        assert end_size < block_size

        # we read at most block_size bytes
        bytes_num_to_read = min(block_size, seekable.size)

        if beg_size > 0:
            beg_content = seekable.read_at(0, bytes_num_to_read).lstrip()
            beg_ints = Magika._get_beg_ints_with_padding(
                beg_content, beg_size, padding_token
            )
        else:
            beg_ints = []

        if mid_size > 0:
            # mid_idx points to the left-most offset to read for the "mid" component
            # of the features.
            mid_bytes_num_to_read = min(seekable.size, mid_size)
            mid_idx = (seekable.size - mid_bytes_num_to_read) // 2
            mid_content = seekable.read_at(mid_idx, mid_bytes_num_to_read)
            mid_ints = Magika._get_mid_ints_with_padding(
                mid_content, mid_size, padding_token
            )
        else:
            mid_ints = []

        if end_size > 0:
            end_content = seekable.read_at(
                seekable.size - bytes_num_to_read, bytes_num_to_read
            ).rstrip()
            end_ints = Magika._get_end_ints_with_padding(
                end_content, end_size, padding_token
            )
        else:
            end_ints = []

        if use_inputs_at_offsets:
            offset_0x8000_0x8007 = Magika._get_ints_at_offset_or_padding(
                seekable, 0x8000, 8, padding_token
            )
            offset_0x8800_0x8807 = Magika._get_ints_at_offset_or_padding(
                seekable, 0x8800, 8, padding_token
            )
            offset_0x9000_0x9007 = Magika._get_ints_at_offset_or_padding(
                seekable, 0x9000, 8, padding_token
            )
            offset_0x9800_0x9807 = Magika._get_ints_at_offset_or_padding(
                seekable, 0x9800, 8, padding_token
            )
        else:
            offset_0x8000_0x8007 = []
            offset_0x8800_0x8807 = []
            offset_0x9000_0x9007 = []
            offset_0x9800_0x9807 = []

        return ModelFeatures(
            beg=beg_ints,
            mid=mid_ints,
            end=end_ints,
            offset_0x8000_0x8007=offset_0x8000_0x8007,
            offset_0x8800_0x8807=offset_0x8800_0x8807,
            offset_0x9000_0x9007=offset_0x9000_0x9007,
            offset_0x9800_0x9807=offset_0x9800_0x9807,
        )

    @staticmethod
    def _get_beg_ints_with_padding(
        beg_content: bytes, beg_size: int, padding_token: int
    ) -> List[int]:
        """Take an (already-stripped) buffer as input and extract beg ints.
        This returns a list of integers whose length is exactly beg_size. If
        the buffer is bigger than required, take only the initial portion. If
        the buffer is shorter, add padding at the end.
        """

        if beg_size < len(beg_content):
            # we don't need so many bytes
            beg_content = beg_content[0:beg_size]

        beg_ints = list(map(int, beg_content))

        if len(beg_ints) < beg_size:
            # we don't have enough ints, add padding
            beg_ints = beg_ints + ([padding_token] * (beg_size - len(beg_ints)))

        assert len(beg_ints) == beg_size

        return beg_ints

    @staticmethod
    def _get_mid_ints_with_padding(
        mid_content: bytes, mid_size: int, padding_token: int
    ) -> List[int]:
        """Take a buffer as input and extract mid ints. This returns a list of
        integers whose length is exactly mid_size. If the buffer is bigger than
        required, take only its middle part. If the buffer is shorter, add
        padding to its left and right. If we need to add an odd number of
        padding integers, add an extra one to the right.
        """

        if mid_size < len(mid_content):
            mid_idx = (len(mid_content) - mid_size) // 2
            mid_content = mid_content[mid_idx : mid_idx + mid_size]

        mid_ints = list(map(int, mid_content))

        if len(mid_ints) < mid_size:
            # we don't have enough ints, add padding
            padding_size = mid_size - len(mid_ints)
            padding_size_left = padding_size // 2
            padding_size_right = padding_size - padding_size_left
            mid_ints = (
                ([padding_token] * padding_size_left)
                + mid_ints
                + ([padding_token] * padding_size_right)
            )

        assert len(mid_ints) == mid_size

        return mid_ints

    @staticmethod
    def _get_end_ints_with_padding(
        end_content: bytes, end_size: int, padding_token: int
    ) -> List[int]:
        """Take an (already-stripped) buffer as input and extract end ints. This
        returns a list of integers whose length is exactly end_size.  If the
        buffer is bigger than required, take only the last portion. If the
        buffer is shorter, add padding at the beginning.
        """

        if end_size < len(end_content):
            # we don't need so many bytes
            end_content = end_content[len(end_content) - end_size : len(end_content)]

        end_ints = list(map(int, end_content))

        if len(end_ints) < end_size:
            # we don't have enough ints, add padding
            end_ints = ([padding_token] * (end_size - len(end_ints))) + end_ints

        assert len(end_ints) == end_size

        return end_ints

    @staticmethod
    def _get_ints_at_offset_or_padding(
        seekable: Seekable, offset: int, size: int, padding_token: int
    ) -> List[int]:
        if offset + size <= seekable.size:
            return list(map(int, seekable.read_at(offset, size)))
        return [padding_token] * size

    def _get_model_outputs_from_features(
        self, all_features: List[Tuple[Path, ModelFeatures]]
    ) -> List[Tuple[Path, ModelOutput]]:
        raw_preds = self._get_raw_predictions(all_features)
        top_preds_idxs = np.argmax(raw_preds, axis=1)
        preds_content_types_labels = self._target_labels_space_np[top_preds_idxs]
        scores = np.max(raw_preds, axis=1)

        return [
            (path, ModelOutput(ct_label=ContentTypeLabel(ct_label), score=float(score)))
            for (path, _), ct_label, score in zip(
                all_features, preds_content_types_labels, scores
            )
        ]

    def _get_results_from_features(
        self, all_features: List[Tuple[Path, ModelFeatures]]
    ) -> Dict[str, MagikaResult]:
        # We now do inference for those files that need it.

        if len(all_features) == 0:
            # nothing to be done
            return {}

        results: Dict[str, MagikaResult] = {}

        for path, model_output in self._get_model_outputs_from_features(all_features):
            # In additional to the content type label from the DL model, we
            # also allow for other logic to overwrite such result. For
            # debugging and information purposes, the JSON output stores
            # both the raw DL model output and the final output we return to
            # the user.

            output_ct_label, overwrite_reason = (
                self._get_output_ct_label_from_dl_result(
                    model_output.ct_label, model_output.score
                )
            )

            results[str(path)] = self._get_result_from_labels_and_score(
                path=path,
                dl_ct_label=model_output.ct_label,
                output_ct_label=output_ct_label,
                score=model_output.score,
                overwrite_reason=overwrite_reason,
            )

        return results

    def _get_result_from_features(
        self, features: ModelFeatures, path: Optional[Path] = None
    ) -> MagikaResult:
        # This is useful to scan from stream of bytes
        if path is None:
            path = Path("-")
        all_features = [(Path("-"), features)]
        result_with_dl = self._get_results_from_features(all_features)["-"]
        return result_with_dl

    def _get_output_ct_label_from_dl_result(
        self, dl_ct_label: ContentTypeLabel, score: float
    ) -> Tuple[ContentTypeLabel, OverwriteReason]:
        overwrite_reason = OverwriteReason.NONE

        # Overwrite dl_ct_label if specified in the overwrite_map model config
        output_ct_label = self._model_config.overwrite_map.get(dl_ct_label, dl_ct_label)
        if output_ct_label != dl_ct_label:
            overwrite_reason = OverwriteReason.OVERWRITE_MAP

        if self._prediction_mode == PredictionMode.BEST_GUESS:
            # We take the (potentially overwritten) model prediction, no matter
            # what the score is.
            pass
        elif (
            self._prediction_mode == PredictionMode.HIGH_CONFIDENCE
            and score
            >= self._model_config.thresholds.get(
                dl_ct_label, self._model_config.medium_confidence_threshold
            )
        ):
            # The model score is higher than the per-content-type
            # high-confidence threshold, so we keep it.
            pass
        elif (
            self._prediction_mode == PredictionMode.MEDIUM_CONFIDENCE
            and score >= self._model_config.medium_confidence_threshold
        ):
            # The model score is higher than the generic medium-confidence
            # threshold, so we keep it.
            pass
        else:
            # We are not in a condition to trust the model, we opt to return
            # generic labels. Note that here we use an implicit assumption that
            # the model has, at the very least, got the binary vs. text category
            # right. This allows us to pick between unknown and txt without the
            # need to read or scan the file bytes once again.
            overwrite_reason = OverwriteReason.LOW_CONFIDENCE
            if self._get_ct_info(output_ct_label).is_text:
                output_ct_label = ContentTypeLabel.TXT
            else:
                output_ct_label = ContentTypeLabel.UNKNOWN

        return output_ct_label, overwrite_reason

    def _get_result_from_labels_and_score(
        self,
        path: Path,
        dl_ct_label: ContentTypeLabel,
        output_ct_label: ContentTypeLabel,
        score: float,
        overwrite_reason: OverwriteReason = OverwriteReason.NONE,
    ) -> MagikaResult:
        return MagikaResult(
            path=path,
            prediction=MagikaPrediction(
                dl=self._get_ct_info(dl_ct_label),
                output=self._get_ct_info(output_ct_label),
                score=score,
                overwrite_reason=overwrite_reason,
            ),
        )

    def _get_result_or_features_from_path(
        self, path: Path
    ) -> Tuple[Optional[MagikaResult], Optional[ModelFeatures]]:
        """
        Given a path, we return either a MagikaOutput or a MagikaFeatures.

        There are some files and corner cases for which we do not need to use
        deep learning to get the output; in these cases, we already return a
        MagikaOutput object.

        For some other files, we do need to use deep learning, in which case we
        return a MagikaFeatures object. Note that for now we just collect the
        features instead of already performing inference because we want to use
        batching.
        """

        if self._no_dereference and path.is_symlink():
            result = self._get_result_from_labels_and_score(
                path=path,
                dl_ct_label=ContentTypeLabel.UNDEFINED,
                output_ct_label=ContentTypeLabel.SYMLINK,
                score=1.0,
            )
            return result, None

        if not path.exists():
            return MagikaResult(path=path, status=Status.FILE_NOT_FOUND_ERROR), None

        if path.is_file():
            if path.stat().st_size == 0:
                result = self._get_result_from_labels_and_score(
                    path=path,
                    dl_ct_label=ContentTypeLabel.UNDEFINED,
                    output_ct_label=ContentTypeLabel.EMPTY,
                    score=1.0,
                )
                return result, None

            elif not os.access(path, os.R_OK):
                return MagikaResult(path=path, status=Status.PERMISSION_ERROR), None

            elif path.stat().st_size <= self._model_config.min_file_size_for_dl:
                result = self._get_result_from_first_block_of_file(path)
                return result, None

            else:
                file_features = Magika._extract_features_from_path(
                    path,
                    self._model_config.beg_size,
                    self._model_config.mid_size,
                    self._model_config.end_size,
                    self._model_config.padding_token,
                    self._model_config.block_size,
                    self._model_config.use_inputs_at_offsets,
                )
                # Check whether we have enough bytes for a meaningful
                # detection, and not just padding.
                if (
                    file_features.beg[self._model_config.min_file_size_for_dl - 1]
                    == self._model_config.padding_token
                ):
                    # If the n-th token is padding, then it means that,
                    # post-stripping, we do not have enough meaningful
                    # bytes.
                    result = self._get_result_from_first_block_of_file(path)
                    return result, None

                else:
                    # We have enough bytes, return the features for a model
                    # prediction.
                    return None, file_features

        elif path.is_dir():
            result = self._get_result_from_labels_and_score(
                path=path,
                dl_ct_label=ContentTypeLabel.UNDEFINED,
                output_ct_label=ContentTypeLabel.DIRECTORY,
                score=1.0,
            )
            return result, None

        else:
            result = self._get_result_from_labels_and_score(
                path=path,
                dl_ct_label=ContentTypeLabel.UNDEFINED,
                output_ct_label=ContentTypeLabel.UNKNOWN,
                score=1.0,
            )
            return result, None

        raise Exception("unreachable")

    def _get_result_or_features_from_bytes(
        self, content: bytes
    ) -> Tuple[Optional[MagikaResult], Optional[ModelFeatures]]:
        if len(content) == 0:
            result = self._get_result_from_labels_and_score(
                path=Path("-"),
                dl_ct_label=ContentTypeLabel.UNDEFINED,
                output_ct_label=ContentTypeLabel.EMPTY,
                score=1.0,
            )
            return result, None

        elif len(content) <= self._model_config.min_file_size_for_dl:
            result = self._get_result_from_few_bytes(content)
            return result, None

        else:
            file_features = Magika._extract_features_from_bytes(
                content,
                self._model_config.beg_size,
                self._model_config.mid_size,
                self._model_config.end_size,
                self._model_config.padding_token,
                self._model_config.block_size,
                self._model_config.use_inputs_at_offsets,
            )
            # Check whether we have enough bytes for a meaningful
            # detection, and not just padding.
            if (
                file_features.beg[self._model_config.min_file_size_for_dl - 1]
                == self._model_config.padding_token
            ):
                # If the n-th token is padding, then it means that,
                # post-stripping, we do not have enough meaningful
                # bytes.
                result = self._get_result_from_few_bytes(content)
                return result, None

            else:
                # We have enough bytes, return the features for a model
                # prediction.
                return None, file_features

        raise Exception("unreachable")

    def _get_result_or_features_from_stream(
        self, stream: BinaryIO
    ) -> Tuple[Optional[MagikaResult], Optional[ModelFeatures]]:
        """
        Given a `BinaryIO` stream, we return either a MagikaOutput or a MagikaFeatures.

        There are some corner cases for which we do not need to use deep
        learning to get the output; in these cases, we return directly a
        MagikaOutput object.

        For all other cases, we do need to use deep learning, in which case we
        return a MagikaFeatures object. Note that for now we just collect the
        features instead of already performing inference because we want to use
        batching.
        """

        stream.seek(0, io.SEEK_END)
        bytes_stream_size = stream.tell()

        if bytes_stream_size == 0:
            result = self._get_result_from_labels_and_score(
                path=Path("-"),
                dl_ct_label=ContentTypeLabel.UNDEFINED,
                output_ct_label=ContentTypeLabel.EMPTY,
                score=1.0,
            )
            return result, None

        elif bytes_stream_size <= self._model_config.min_file_size_for_dl:
            stream.seek(0)
            content = stream.read()
            result = self._get_result_from_few_bytes(content)
            return result, None

        else:
            file_features = Magika._extract_features_from_stream(
                stream,
                self._model_config.beg_size,
                self._model_config.mid_size,
                self._model_config.end_size,
                self._model_config.padding_token,
                self._model_config.block_size,
                self._model_config.use_inputs_at_offsets,
            )
            # Check whether we have enough bytes for a meaningful
            # detection, and not just padding.
            if (
                file_features.beg[self._model_config.min_file_size_for_dl - 1]
                == self._model_config.padding_token
            ):
                # If the n-th token is padding, then it means that,
                # post-stripping, we do not have enough meaningful
                # bytes.
                stream.seek(0)
                content = stream.read()
                result = self._get_result_from_few_bytes(content)
                return result, None

            else:
                # We have enough bytes, return the features for a model
                # prediction.
                return None, file_features

        raise Exception("unreachable")

    def _get_result_from_first_block_of_file(self, path: Path) -> MagikaResult:
        # We read at most "block_size" bytes
        with open(path, "rb") as f:
            content = f.read(self._model_config.block_size)
        return self._get_result_from_few_bytes(content, path)

    def _get_result_from_few_bytes(
        self, content: bytes, path: Path = Path("-")
    ) -> MagikaResult:
        assert len(content) <= 4 * self._model_config.block_size
        ct_label = self._get_ct_label_from_few_bytes(content)
        return self._get_result_from_labels_and_score(
            path=path,
            dl_ct_label=ContentTypeLabel.UNDEFINED,
            output_ct_label=ct_label,
            score=1.0,
        )

    def _get_ct_label_from_few_bytes(self, content: bytes) -> ContentTypeLabel:
        try:
            ct_label = ContentTypeLabel.TXT
            _ = content.decode("utf-8")
        except UnicodeDecodeError:
            ct_label = ContentTypeLabel.UNKNOWN
        return ct_label

    def _get_raw_predictions(
        self, features: List[Tuple[Path, ModelFeatures]]
    ) -> npt.NDArray:
        """
        Given a list of (path, features), return a (files_num, features_size)
        matrix encoding the predictions.
        """

        start_time = time.time()
        X_bytes = []
        for _, fs in features:
            sample_bytes = []
            if self._model_config.beg_size > 0:
                sample_bytes.extend(fs.beg[: self._model_config.beg_size])
            if self._model_config.mid_size > 0:
                sample_bytes.extend(fs.mid[: self._model_config.mid_size])
            if self._model_config.end_size > 0:
                sample_bytes.extend(fs.end[-self._model_config.end_size :])
            X_bytes.append(sample_bytes)
        X = np.array(X_bytes, dtype=np.int32)
        elapsed_time = 1000 * (time.time() - start_time)
        self._log.debug(f"DL input prepared in {elapsed_time:.03f} ms")

        raw_predictions_list = []
        samples_num = X.shape[0]

        max_internal_batch_size = 1000
        batches_num = samples_num // max_internal_batch_size
        if samples_num % max_internal_batch_size != 0:
            batches_num += 1

        for batch_idx in range(batches_num):
            self._log.debug(
                f"Getting raw predictions for (internal) batch {batch_idx+1}/{batches_num}"
            )
            start_idx = batch_idx * max_internal_batch_size
            end_idx = min((batch_idx + 1) * max_internal_batch_size, samples_num)

            start_time = time.time()
            batch_raw_predictions = self._onnx_session.run(
                ["target_label"], {"bytes": X[start_idx:end_idx, :]}
            )[0]
            elapsed_time = 1000 * (time.time() - start_time)
            self._log.debug(f"DL raw prediction in {elapsed_time:.03f} ms")

            raw_predictions_list.append(batch_raw_predictions)
        return np.concatenate(raw_predictions_list)



#!/usr/bin/env python3

import json
import shutil
import sys
from pathlib import Path
from typing import Optional

import click

MODELS_NAMES_TO_INCLUDE_IN_PACKAEG = [
    "standard_v3_2",
]


REPO_ROOT_DIR = Path(__file__).parent.parent.parent
assert REPO_ROOT_DIR.is_dir() and (REPO_ROOT_DIR / ".git").is_dir()

ASSETS_DIR = REPO_ROOT_DIR / "assets"
assert ASSETS_DIR.is_dir()

CONTENT_TYPES_KB_PATH = ASSETS_DIR / "content_types_kb.min.json"
assert CONTENT_TYPES_KB_PATH.is_file()

ASSETS_MODELS_DIR = ASSETS_DIR / "models"
assert ASSETS_MODELS_DIR.is_dir()

PYTHON_ROOT_DIR = REPO_ROOT_DIR / "python"
assert PYTHON_ROOT_DIR.is_dir()

PYTHON_CONTENT_TYPES_KB_PATH = (
    PYTHON_ROOT_DIR / "src" / "magika" / "config" / "content_types_kb.min.json"
)

PYTHON_MODELS_DIR = PYTHON_ROOT_DIR / "src" / "magika" / "models"
assert PYTHON_MODELS_DIR.is_dir()

PYTHON_CONTENT_TYPES_LABELS_PY_PATH = (
    PYTHON_ROOT_DIR / "src" / "magika" / "types" / "content_type_label.py"
)


@click.command()
@click.option(
    "--models-names",
    "models_names_str",
    help="Comma-separated list of models names to import in the package",
)
def main(models_names_str: Optional[str]) -> None:
    if models_names_str is None:
        models_names = MODELS_NAMES_TO_INCLUDE_IN_PACKAEG
    else:
        models_names = list(map(lambda s: s.strip(), models_names_str.split(",")))

    print(f"Including these models in the package: {models_names}")

    update_content_type_kb()
    update_content_type_label_py()

    print(f"Deleting {PYTHON_MODELS_DIR}")
    shutil.rmtree(PYTHON_MODELS_DIR)
    for model_name in models_names:
        add_model_to_python_package(model_name)


def update_content_type_kb() -> None:
    print(
        f"Syncing python's content types KB: {CONTENT_TYPES_KB_PATH} => {PYTHON_CONTENT_TYPES_KB_PATH}"
    )
    PYTHON_CONTENT_TYPES_KB_PATH.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(CONTENT_TYPES_KB_PATH, PYTHON_CONTENT_TYPES_KB_PATH)


def add_model_to_python_package(model_name: str) -> None:
    assets_model_dir = ASSETS_MODELS_DIR / model_name
    if not assets_model_dir.is_dir():
        print(f'ERROR: model "{model_name} not found')
        sys.exit(1)

    python_model_dir = PYTHON_MODELS_DIR / model_name

    print(f"Adding model {assets_model_dir} => {python_model_dir}")
    shutil.copytree(assets_model_dir, python_model_dir)


CONTENT_TYPE_LABEL_SOURCE_PREFIX = """
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from magika.types.strenum import StrEnum

# NOTE: DO NOT EDIT --- This file is automatically generated.


# This is the list of all possible content types we know about; however, models
# support a smaller subset of them. See model's README.md for details.
class ContentTypeLabel(StrEnum):
"""


def update_content_type_label_py() -> None:
    print(f"Updating {PYTHON_CONTENT_TYPES_LABELS_PY_PATH}")

    kb = json.loads(CONTENT_TYPES_KB_PATH.read_text())

    enum_body_lines = []
    for ct_label_str in sorted(kb.keys()):
        if ct_label_str[0].isdigit():
            line = (" " * 4) + f'_{ct_label_str.upper()} = "{ct_label_str}"'
        else:
            line = (" " * 4) + f'{ct_label_str.upper()} = "{ct_label_str}"'
        enum_body_lines.append(line)

    out = (
        CONTENT_TYPE_LABEL_SOURCE_PREFIX.strip()
        + "\n"
        + "\n".join(enum_body_lines)
        + "\n"
    )
    out += (
        "\n"
        + (" " * 4)
        + (
            """
    def __repr__(self) -> str:
        return str(self)
    """.strip()
            + "\n"
        )
    )

    PYTHON_CONTENT_TYPES_LABELS_PY_PATH.write_text(out)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2023-2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import re
import subprocess
from pathlib import Path


# TODO(https://github.com/PyO3/maturin/issues/2163): Remove this file when fixed.
def main() -> None:
    repo_root_dir = Path(__file__).parent.parent.parent
    python_root_dir = repo_root_dir / "python"
    rust_root_dir = repo_root_dir / "rust"
    # Compute paths to files we'll need to restore at the end of the build
    rust_main_rs_path = rust_root_dir / "cli" / "src" / "main.rs"
    rust_cli_cargo_toml_path = rust_root_dir / "cli" / "Cargo.toml"

    # get the rust version from Cargo.toml and patch main.rs
    rust_version = get_rust_version(rust_root_dir)
    patch_main_rs_with_version(rust_main_rs_path, rust_version)

    # get the python version from magika.__version__ and patch Cargo.toml
    python_version = get_python_version(python_root_dir)
    patch_cargo_toml_with_version(rust_cli_cargo_toml_path, python_version)

    # update Cargo.lock
    subprocess.run(["cargo", "check"], cwd=rust_root_dir / "cli", check=True)


def get_rust_version(rust_root_dir: Path) -> str:
    cargo_path = rust_root_dir / "cli" / "Cargo.toml"
    version = extract_with_regex(cargo_path, 'version = "([A-Za-z0-9.-]+)".*')
    print(f"Extracted rust version: {version}")
    return version


def get_python_version(python_root_dir: Path) -> str:
    init_path = python_root_dir / "src" / "magika" / "__init__.py"
    version = extract_with_regex(init_path, '__version__ = "([A-Za-z0-9.-]+)"')
    print(f"Extracted python version: {version}")
    return version


def patch_main_rs_with_version(rust_main_rs_path: Path, version: str) -> None:
    print(f'Patching {rust_main_rs_path} with rust version "{version}"')
    patch_line_matching_prefix(
        rust_main_rs_path,
        "        let binary = clap::crate_version!();",
        f'        let binary = "{version}";',
    )


def patch_cargo_toml_with_version(cargo_toml_path: Path, version: str) -> None:
    print(f'Patching {cargo_toml_path} with python version "{version}"')
    patch_line_matching_prefix(cargo_toml_path, "version = ", f'version = "{version}"')


def extract_with_regex(file_path: Path, regex: str) -> str:
    """Extract a string via regex. This raises an exception if no or more than
    one matches are found."""

    lines = file_path.read_text().split("\n")
    output = None
    for line in lines:
        m = re.fullmatch(regex, line)
        if m:
            if output is not None:
                raise Exception(
                    f'ERROR: Found more than one match for "{regex}" in {file_path}'
                )
            output = m.group(1)
    if output is None:
        raise Exception(f'No hits for "{regex}" in {file_path}')
    return output


def patch_line_matching_prefix(file_path: Path, prefix: str, new_line: str) -> None:
    """Patch a line starting with a given prefix with a new line. This raises an
    exception if no such line or more than one lines with a given prefix are
    found."""

    lines = file_path.read_text().split("\n")
    already_found = False
    for line_idx in range(len(lines)):
        line = lines[line_idx]
        if line.startswith(prefix):
            if already_found:
                raise Exception(
                    f'ERROR: Found more than one line with prefix "{prefix}" in "{file_path}"'
                )
            already_found = True
            lines[line_idx] = new_line

    if not already_found:
        raise Exception(
            f'ERROR: Did not find any line with prefix "{prefix}" in "{file_path}"'
        )
    file_path.write_text("\n".join(lines))


if __name__ == "__main__":
    main()



"""Performs a number of sanity checks on Magika's documentation."""

from __future__ import annotations

import re
import sys
from dataclasses import dataclass
from pathlib import Path

import click
import requests

REPO_ROOT_DIR = Path(__file__).parent.parent.parent
assert REPO_ROOT_DIR.is_dir() and (REPO_ROOT_DIR / ".git").is_dir()


IGNORE_PREFIX_PATTERNS = [
    ".mypy_cache",
    ".pytest_cache",
    ".ruff_cache",
    "python/.venv",
    "python/dist",
]

# List of (prefix) of urls that return non-200 even if they are valid; for
# simplicity, we just skip checking them.
URLS_ALLOWLIST_PREFIXES = [
    "https://api.securityscorecards.dev/projects/github.com/google/magika/badge",
    "https://crates.io/crates/magika",
    "https://crates.io/crates/magika-cli",
    "https://www.unrealengine.com/en-US",
    "https://www.unrealengine.com/marketplace/en-US/store",
    "https://www.virustotal.com/",
]


@click.command()
@click.option("--skip-external-validity-check", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
def main(skip_external_validity_check: bool, verbose: bool) -> None:
    with_errors = False

    success = check_versions_are_up_to_date()
    if not success:
        with_errors = True

    success = check_markdown_links(skip_external_validity_check, verbose)
    if not success:
        with_errors = True

    if with_errors:
        print("There was at least one error.")
        sys.exit(1)

    print("Everything looks good.")


def check_versions_are_up_to_date() -> bool:
    """Checks that the mentioned latest versions and models are up to date.
    Returns True if everything is good, False otherwise."""

    # Check that the versions mentioned in the READMEs are up to date
    python_latest_stable_version = get_python_latest_stable_version()
    python_default_model_name = get_python_default_model_name()

    print(f"INFO: {python_latest_stable_version=} {python_default_model_name=}")

    expected_lines = [
        f"> - The documentation on GitHub refers to the latest, potentially unreleased and unstable version of Magika. The latest stable release of the `magika` Python package is `{python_latest_stable_version}`, and you can consult the associated documenation [here](https://github.com/google/magika/tree/python-v{python_latest_stable_version}). You can install the latest stable version with: `pip install magika`.",
        f"- Trained and evaluated on a dataset of ~100M files across [200+ content types](./assets/models/{python_default_model_name}/README.md).",
        f"- [List of supported content types by the latest model, `{python_default_model_name}`](./assets/models/{python_default_model_name}/README.md)",
        f"| [Python `Magika` module](./python/README.md) | Stable enough for prod use cases | [`{python_default_model_name}`](./assets/models/{python_default_model_name}/README.md) |",
    ]

    readme_content_lines_set = set(
        (REPO_ROOT_DIR / "README.md").read_text().split("\n")
    )

    with_errors = False
    for expected_line in expected_lines:
        if expected_line not in readme_content_lines_set:
            print(f'ERROR: could not find the following line: "{expected_line}"')
            with_errors = True

    success = with_errors is False
    return success


def get_python_latest_stable_version() -> str:
    res = requests.get("https://pypi.org/pypi/magika/json")
    assert res.status_code == 200
    latest_stable_version = res.json().get("info", {}).get("version", None)
    assert latest_stable_version is not None
    return latest_stable_version


def get_python_default_model_name() -> str:
    default_model_name = None
    magika_path = REPO_ROOT_DIR / "python" / "src" / "magika" / "magika.py"
    assert magika_path.is_file()
    for line in magika_path.read_text().split("\n"):
        m = re.fullmatch('_DEFAULT_MODEL_NAME = "([a-zA-Z0-9_]+)"', line)
        if m:
            # If we already found something, there is a bug somewhere
            assert default_model_name is None
            default_model_name = m.group(1)

    return default_model_name


def check_markdown_links(skip_external_validity_check: bool, verbose: bool) -> bool:
    """Checks that links in Markdown files are OK. Returns True if everything is
    good, False otherwise."""

    with_errors = False
    for path in enumerate_markdown_files_in_dir(Path(".")):
        if verbose:
            print(f"Analyzing {path}")
        for ui in extract_uris_infos_from_file(
            path,
            skip_external_validity_check=skip_external_validity_check,
            verbose=verbose,
        ):
            if not ui.is_valid:
                with_errors = True
                print(
                    f"ERROR: {path.relative_to(REPO_ROOT_DIR)} has non-valid uri: {ui.uri}"
                )

            # For python/README.md (which is used on pypi), we also check that
            # the URIs are either pointing to an external resource or are pure
            # anchors.
            if str(path.relative_to(REPO_ROOT_DIR)) == "python/README.md":
                if not ui.is_external and not ui.is_pure_anchor:
                    with_errors = True
                    print(
                        f"ERROR: {path.relative_to(REPO_ROOT_DIR)}, in python/, has a non-external uri: {ui.uri}"
                    )

    success = with_errors is False
    return success


def enumerate_markdown_files_in_dir(rel_dir: Path) -> list[Path]:
    if rel_dir.is_absolute():
        print(f"{rel_dir} is not relative")
        sys.exit(1)
    a_dir = REPO_ROOT_DIR / rel_dir
    assert a_dir.is_dir()
    paths: list[Path] = []
    for path in sorted(a_dir.rglob("*.md")):
        should_ignore = False
        for exclude_prefix_pattern in IGNORE_PREFIX_PATTERNS:
            if str(path.relative_to(REPO_ROOT_DIR)).startswith(exclude_prefix_pattern):
                should_ignore = True
                break
        if not should_ignore:
            paths.append(path)
    return paths


def extract_uris_infos_from_file(
    path: Path, skip_external_validity_check: bool, verbose: bool
) -> list[UriInfo]:
    uri_regex = r"\[.*?\]\((.*?)\)"
    uris: list[str] = re.findall(uri_regex, path.read_text())

    uris_infos: list[UriInfo] = []
    for uri in uris:
        if verbose:
            print(f"Analyzing uri: {uri}")

        is_external = uri.startswith("http://") or uri.startswith("https://")
        is_valid = None
        is_pure_anchor = None
        is_insecure = None

        if is_external:
            is_pure_anchor = False
            if uri.startswith("http://"):
                is_insecure = True
                print(f"WARNING: {uri} is not using https")
            else:
                is_insecure = False

            if skip_external_validity_check:
                print(f"WARNING: skipping check for {uri}")
                is_valid = True
            else:
                for url_allowlist_prefix in URLS_ALLOWLIST_PREFIXES:
                    if uri.startswith(url_allowlist_prefix):
                        is_valid = True
                        break
                if is_valid is not True:
                    r = requests.head(uri, allow_redirects=True, timeout=5)
                    if r.status_code == 200:
                        is_valid = True
                    else:
                        print(
                            f"ERROR: Got non-200 response status code for {uri}: {r.status_code}"
                        )
                        is_valid = False
        else:
            is_insecure = False
            if uri.startswith("#"):
                is_valid = True
                is_pure_anchor = True
            else:
                is_pure_anchor = False
                if Path(uri).is_absolute():
                    is_valid = False
                else:
                    if uri.find("#") >= 0:
                        # This URI is not a pure anchor, but it does have an
                        # anchor. We remove it so that we can check whether the
                        # file exists or not.
                        rel_file_path = uri.split("#")[0]
                    else:
                        rel_file_path = uri
                    abs_path = path.parent / rel_file_path
                    is_valid = abs_path.is_file() or abs_path.is_dir()

        assert is_valid is not None
        assert is_pure_anchor is not None
        assert is_insecure is not None

        uris_infos.append(
            UriInfo(
                uri=uri,
                is_external=is_external,
                is_valid=is_valid,
                is_pure_anchor=is_pure_anchor,
                is_insecure=is_insecure,
            )
        )

    return uris_infos


@dataclass(kw_only=True)
class UriInfo:
    uri: str
    is_external: bool
    is_valid: bool
    is_pure_anchor: bool
    is_insecure: bool


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from magika.types.strenum import StrEnum


class Status(StrEnum):
    OK = "ok"

    # Used when a file path does not exist
    FILE_NOT_FOUND_ERROR = "file_not_found_error"

    # Used when a file path exists, but there are permission issues, e.g., can't
    # read file
    PERMISSION_ERROR = "permission_error"

    # Represents a generic error-like unknown status.
    UNKNOWN = "unknown"



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from magika.types.strenum import StrEnum

# NOTE: DO NOT EDIT --- This file is automatically generated.


# This is the list of all possible content types we know about; however, models
# support a smaller subset of them. See model's README.md for details.
class ContentTypeLabel(StrEnum):
    _3DS = "3ds"
    _3DSM = "3dsm"
    _3DSX = "3dsx"
    _3GP = "3gp"
    _3MF = "3mf"
    ABNF = "abnf"
    ACE = "ace"
    ADA = "ada"
    AFF = "aff"
    AI = "ai"
    AIDL = "aidl"
    ALGOL68 = "algol68"
    ANI = "ani"
    APK = "apk"
    APPLEBPLIST = "applebplist"
    APPLEDOUBLE = "appledouble"
    APPLEPLIST = "appleplist"
    APPLESINGLE = "applesingle"
    AR = "ar"
    ARC = "arc"
    ARJ = "arj"
    ARROW = "arrow"
    ASC = "asc"
    ASD = "asd"
    ASF = "asf"
    ASM = "asm"
    ASP = "asp"
    AU = "au"
    AUTOHOTKEY = "autohotkey"
    AUTOIT = "autoit"
    AVI = "avi"
    AVIF = "avif"
    AVRO = "avro"
    AWK = "awk"
    AX = "ax"
    BATCH = "batch"
    BAZEL = "bazel"
    BCAD = "bcad"
    BIB = "bib"
    BMP = "bmp"
    BPG = "bpg"
    BPL = "bpl"
    BRAINFUCK = "brainfuck"
    BRF = "brf"
    BZIP = "bzip"
    BZIP3 = "bzip3"
    C = "c"
    CAB = "cab"
    CAD = "cad"
    CAT = "cat"
    CDF = "cdf"
    CHM = "chm"
    CLOJURE = "clojure"
    CMAKE = "cmake"
    COBOL = "cobol"
    COFF = "coff"
    COFFEESCRIPT = "coffeescript"
    COM = "com"
    CPL = "cpl"
    CPP = "cpp"
    CRT = "crt"
    CRX = "crx"
    CS = "cs"
    CSPROJ = "csproj"
    CSS = "css"
    CSV = "csv"
    CTL = "ctl"
    DART = "dart"
    DEB = "deb"
    DEX = "dex"
    DEY = "dey"
    DICOM = "dicom"
    DIFF = "diff"
    DIRECTORY = "directory"
    DJANGO = "django"
    DLL = "dll"
    DM = "dm"
    DMG = "dmg"
    DMIGD = "dmigd"
    DMSCRIPT = "dmscript"
    DOC = "doc"
    DOCKERFILE = "dockerfile"
    DOCX = "docx"
    DOSMBR = "dosmbr"
    DOTX = "dotx"
    DSSTORE = "dsstore"
    DWG = "dwg"
    DXF = "dxf"
    DYLIB = "dylib"
    EBML = "ebml"
    ELF = "elf"
    ELIXIR = "elixir"
    EMF = "emf"
    EML = "eml"
    EMPTY = "empty"
    EPUB = "epub"
    ERB = "erb"
    ERLANG = "erlang"
    ESE = "ese"
    EXE = "exe"
    EXP = "exp"
    FLAC = "flac"
    FLUTTER = "flutter"
    FLV = "flv"
    FORTRAN = "fortran"
    FPX = "fpx"
    GEMFILE = "gemfile"
    GEMSPEC = "gemspec"
    GIF = "gif"
    GITATTRIBUTES = "gitattributes"
    GITMODULES = "gitmodules"
    GLEAM = "gleam"
    GO = "go"
    GPX = "gpx"
    GRADLE = "gradle"
    GROOVY = "groovy"
    GZIP = "gzip"
    H = "h"
    H5 = "h5"
    HANDLEBARS = "handlebars"
    HASKELL = "haskell"
    HCL = "hcl"
    HEIF = "heif"
    HFS = "hfs"
    HLP = "hlp"
    HPP = "hpp"
    HTA = "hta"
    HTACCESS = "htaccess"
    HTML = "html"
    HVE = "hve"
    HWP = "hwp"
    ICC = "icc"
    ICNS = "icns"
    ICO = "ico"
    ICS = "ics"
    IGNOREFILE = "ignorefile"
    IMG = "img"
    INI = "ini"
    INTERNETSHORTCUT = "internetshortcut"
    IOSAPP = "iosapp"
    IPYNB = "ipynb"
    ISO = "iso"
    JAR = "jar"
    JAVA = "java"
    JAVABYTECODE = "javabytecode"
    JAVASCRIPT = "javascript"
    JINJA = "jinja"
    JNG = "jng"
    JNLP = "jnlp"
    JP2 = "jp2"
    JPEG = "jpeg"
    JSON = "json"
    JSONC = "jsonc"
    JSONL = "jsonl"
    JSX = "jsx"
    JULIA = "julia"
    JXL = "jxl"
    KO = "ko"
    KOTLIN = "kotlin"
    KS = "ks"
    LATEX = "latex"
    LATEXAUX = "latexaux"
    LESS = "less"
    LHA = "lha"
    LICENSE = "license"
    LISP = "lisp"
    LITCS = "litcs"
    LNK = "lnk"
    LOCK = "lock"
    LRZ = "lrz"
    LUA = "lua"
    LZ = "lz"
    LZ4 = "lz4"
    LZX = "lzx"
    M3U = "m3u"
    M4 = "m4"
    MACHO = "macho"
    MAFF = "maff"
    MAKEFILE = "makefile"
    MARKDOWN = "markdown"
    MATLAB = "matlab"
    MHT = "mht"
    MIDI = "midi"
    MKV = "mkv"
    MP2 = "mp2"
    MP3 = "mp3"
    MP4 = "mp4"
    MPEGTS = "mpegts"
    MSCOMPRESS = "mscompress"
    MSI = "msi"
    MSIX = "msix"
    MST = "mst"
    MUI = "mui"
    MUM = "mum"
    MUN = "mun"
    NIM = "nim"
    NPY = "npy"
    NPZ = "npz"
    NULL = "null"
    NUPKG = "nupkg"
    OBJECT = "object"
    OBJECTIVEC = "objectivec"
    OCAML = "ocaml"
    OCX = "ocx"
    ODEX = "odex"
    ODIN = "odin"
    ODP = "odp"
    ODS = "ods"
    ODT = "odt"
    OGG = "ogg"
    OLE = "ole"
    ONE = "one"
    ONNX = "onnx"
    OOXML = "ooxml"
    OTF = "otf"
    OUTLOOK = "outlook"
    PALMOS = "palmos"
    PARQUET = "parquet"
    PASCAL = "pascal"
    PBM = "pbm"
    PCAP = "pcap"
    PDB = "pdb"
    PDF = "pdf"
    PEBIN = "pebin"
    PEM = "pem"
    PERL = "perl"
    PGP = "pgp"
    PHP = "php"
    PICKLE = "pickle"
    PNG = "png"
    PO = "po"
    POSTSCRIPT = "postscript"
    POWERSHELL = "powershell"
    PPT = "ppt"
    PPTX = "pptx"
    PRINTFOX = "printfox"
    PROLOG = "prolog"
    PROTEINDB = "proteindb"
    PROTO = "proto"
    PROTOBUF = "protobuf"
    PSD = "psd"
    PUB = "pub"
    PYTHON = "python"
    PYTHONBYTECODE = "pythonbytecode"
    PYTHONPAR = "pythonpar"
    PYTORCH = "pytorch"
    QOI = "qoi"
    QT = "qt"
    R = "r"
    RANDOMASCII = "randomascii"
    RANDOMBYTES = "randombytes"
    RANDOMTXT = "randomtxt"
    RAR = "rar"
    RDF = "rdf"
    RDP = "rdp"
    RIFF = "riff"
    RLIB = "rlib"
    RLL = "rll"
    RPM = "rpm"
    RST = "rst"
    RTF = "rtf"
    RUBY = "ruby"
    RUST = "rust"
    RZIP = "rzip"
    SCALA = "scala"
    SCHEME = "scheme"
    SCR = "scr"
    SCRIPTWSF = "scriptwsf"
    SCSS = "scss"
    SEVENZIP = "sevenzip"
    SGML = "sgml"
    SH3D = "sh3d"
    SHELL = "shell"
    SMALI = "smali"
    SNAP = "snap"
    SO = "so"
    SOLIDITY = "solidity"
    SQL = "sql"
    SQLITE = "sqlite"
    SQUASHFS = "squashfs"
    SRT = "srt"
    STLBINARY = "stlbinary"
    STLTEXT = "stltext"
    SUM = "sum"
    SVD = "svd"
    SVG = "svg"
    SWF = "swf"
    SWIFT = "swift"
    SYMLINK = "symlink"
    SYMLINKTEXT = "symlinktext"
    SYS = "sys"
    TAR = "tar"
    TCL = "tcl"
    TEXTPROTO = "textproto"
    TGA = "tga"
    THUMBSDB = "thumbsdb"
    TIFF = "tiff"
    TMDX = "tmdx"
    TOML = "toml"
    TORRENT = "torrent"
    TROFF = "troff"
    TSV = "tsv"
    TSX = "tsx"
    TTF = "ttf"
    TWIG = "twig"
    TXT = "txt"
    TXTASCII = "txtascii"
    TXTUTF16 = "txtutf16"
    TXTUTF8 = "txtutf8"
    TYPESCRIPT = "typescript"
    UDF = "udf"
    UNDEFINED = "undefined"
    UNIXCOMPRESS = "unixcompress"
    UNKNOWN = "unknown"
    VBA = "vba"
    VBE = "vbe"
    VCARD = "vcard"
    VCS = "vcs"
    VCXPROJ = "vcxproj"
    VERILOG = "verilog"
    VHD = "vhd"
    VHDL = "vhdl"
    VISIO = "visio"
    VTT = "vtt"
    VUE = "vue"
    WAD = "wad"
    WASM = "wasm"
    WAV = "wav"
    WEBM = "webm"
    WEBP = "webp"
    WEBTEMPLATE = "webtemplate"
    WIM = "wim"
    WINREGISTRY = "winregistry"
    WMA = "wma"
    WMF = "wmf"
    WMV = "wmv"
    WOFF = "woff"
    WOFF2 = "woff2"
    XAR = "xar"
    XCF = "xcf"
    XLS = "xls"
    XLSB = "xlsb"
    XLSX = "xlsx"
    XML = "xml"
    XPI = "xpi"
    XSD = "xsd"
    XZ = "xz"
    YAML = "yaml"
    YARA = "yara"
    ZIG = "zig"
    ZIP = "zip"
    ZLIBSTREAM = "zlibstream"
    ZST = "zst"

    def __repr__(self) -> str:
        return str(self)



"""
We use a StrEnum backport instead of relying on the newly introduced StrEnum
as we want to support at least python 3.8; StrEnum was introduced in python
3.11.

The following code has been taken (and adapted) from:
https://github.com/irgeek/StrEnum/blob/master/strenum/__init__.py#L21
"""

from __future__ import annotations

import enum
from typing import Union


class StrEnum(str, enum.Enum):
    """
    StrEnum is a Python ``enum.Enum`` that inherits from ``str``. The default
    ``auto()`` behavior uses the member name as its value.

    Example usage::

        class Example(StrEnum):
            UPPER_CASE = auto()
            lower_case = auto()
            MixedCase = auto()


        assert Example.UPPER_CASE == "UPPER_CASE"
        assert Example.lower_case == "lower_case"
        assert Example.MixedCase == "MixedCase"
    """

    def __new__(cls, value: Union[str, StrEnum], *args, **kwargs):  # type: ignore[no-untyped-def]
        if not isinstance(value, (str, enum.auto)):
            raise TypeError(
                f"Values of StrEnums must be strings: {value!r} is a {type(value)}"
            )
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self) -> str:
        return str(self.value)

    def _generate_next_value_(name, *_):  # type: ignore[no-untyped-def,override]
        return name


class LowerCaseStrEnum(StrEnum):
    def _generate_next_value_(name, *_):  # type: ignore[no-untyped-def,override]
        return name.lower().replace("_", "-")



#!/usr/bin/env python3
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This script tests a given model against the tests cases, check whether the
predictions are correct, and raise an error otherwise.

This is useful when evaluating new models.
"""

import sys
from pathlib import Path

import click

import magika
from magika import colors


@click.command()
@click.argument("model_dir_or_name")
def main(model_dir_or_name: str) -> None:
    if Path(model_dir_or_name).is_dir():
        model_dir = Path(model_dir_or_name)
    else:
        models_dir = Path(magika.__file__).parent / "models"
        model_dir = models_dir / model_dir_or_name

    if not model_dir.is_dir():
        log_error(f"{model_dir_or_name} is not a dir nor a model name")
        sys.exit(1)

    m = magika.Magika(model_dir=model_dir)

    with_error = False

    tests_data_dir = Path(__file__).parent.parent.parent / "tests_data"
    tests_dirs_names = ["basic", "previous_missdetections"]

    for tests_dir_name in tests_dirs_names:
        tests_dir = tests_data_dir / tests_dir_name
        for test_path in tests_dir.rglob("*"):
            if not test_path.is_file():
                continue

            mr = m.identify_path(test_path)
            assert mr.ok

            prediceted_content_type = mr.prediction.output.label
            expected_content_type = test_path.parent.name
            if prediceted_content_type != expected_content_type:
                with_error = True
                log_error(
                    f'{test_path} predicted as "{prediceted_content_type}" (score: {mr.prediction.score:.4f}), expected "{expected_content_type}".'
                )

    if with_error:
        log_error("There was at least one error.")
    else:
        log_ok("All tests examples were predicted correctly.")


def log_ok(msg: str) -> None:
    print(f"{colors.GREEN}{msg}{colors.RESET}")


def log_error(msg: str) -> None:
    print(f"{colors.RED}ERROR: {msg}{colors.RESET}")


if __name__ == "__main__":
    main()



import warnings
from dataclasses import dataclass
from typing import List

from magika.logger import get_logger
from magika.types.content_type_label import ContentTypeLabel

warnings.simplefilter("always", DeprecationWarning)


@dataclass(frozen=True)
class ContentTypeInfo:
    label: ContentTypeLabel
    mime_type: str
    group: str
    description: str
    extensions: List[str]
    is_text: bool

    @property
    def ct_label(self) -> str:
        warnings.warn(
            "`.ct_label` is deprecated and will be removed in a future version. Use `.label` instead. Consult the documentation for more information.",
            category=DeprecationWarning,
        )
        return str(self.label)

    @property
    def score(self) -> float:
        error_msg = "Unsupported field error: `.score.` is not stored anymore in the `dl` or `output` objects; it is now stored in `MagikaResult`. Consult the documentation for more information."
        log = get_logger()
        log.error(error_msg)
        raise AttributeError(error_msg)

    @property
    def magic(self) -> str:
        warnings.warn(
            "`.magic` is deprecated and will be removed in a future version. Use `.description` instead. Consult the documentation for more information.",
            category=DeprecationWarning,
        )
        return self.description



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import gzip
import json
import math
import random
import string
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import List, Tuple

from magika import Magika
from magika.seekable import Buffer
from magika.types import ModelFeatures
from tests.utils import get_tests_data_dir

random.seed(42)


@dataclass
class TestInfo:
    beg_size: int
    mid_size: int
    end_size: int
    block_size: int
    padding_token: int
    core_content_size: int
    left_ws_num: int
    right_ws_num: int

    __test__ = False


def test_features_extraction_v2(debug: bool = False) -> None:
    tests_cases = _get_tests_cases_from_reference()

    for test_case in tests_cases:
        test_info = TestInfo(**test_case["test_info"])
        test_content = base64.b64decode(test_case["content"])
        expected_features = ModelFeatures(**test_case["features_v2"])

        s = Buffer(test_content)
        features = Magika._extract_features_from_seekable(
            s,
            beg_size=test_info.beg_size,
            mid_size=test_info.mid_size,
            end_size=test_info.end_size,
            padding_token=test_info.padding_token,
            block_size=test_info.block_size,
            use_inputs_at_offsets=True,
        )

        with_error = False
        if features.beg != expected_features.beg:
            with_error = True
            if debug:
                print("beg does not match")
        if features.mid != expected_features.mid:
            with_error = True
            if debug:
                print("mid does not match")
        if features.end != expected_features.end:
            with_error = True
            if debug:
                print("end does not match")
        try:
            assert expected_features == features
        except AssertionError:
            with_error = True
            if debug:
                print("other fields do not match")

        if with_error:
            raise Exception


def _generate_content(test_info: TestInfo) -> bytes:
    """Generate content with a given "core size", with n left and right
    whitespaces, and the core content. with_ws_near_beg and with_ws_near_end
    specify if we need to put spaces near the beg/end, e.g., "A AAA". This is
    useful to test that we don't strip whitespaces that we are not supposed to
    strip."""

    content = _generate_pattern(test_info.core_content_size)

    if test_info.core_content_size >= 5:
        # inject characters that other implementations may mistakenly strip
        content[0] = ord("\x00")
        content[1] = ord(" ")
        content[-2] = ord(" ")
        content[-1] = ord("\x00")

    return (
        _generate_whitespaces(test_info.left_ws_num)
        + bytes(content)
        + _generate_whitespaces(test_info.right_ws_num)
    )


def _generate_whitespaces(size: int) -> bytes:
    whitespaces = string.whitespace
    ws_len = len(whitespaces)
    return bytes([ord(whitespaces[idx % ws_len]) for idx in range(size)])


def _generate_pattern(size: int) -> bytearray:
    """Generate a pattern we can use to test features extraction"""

    chars = string.printable[: 10 + 26 * 2]

    base_pattern = bytearray(chars.encode("ascii"))
    base_pattern_len = len(base_pattern)
    pattern = (base_pattern * int(math.ceil(size / base_pattern_len)))[:size]
    return pattern


def generate_features_extraction_reference():
    beg_size = 512
    mid_size = 512
    end_size = 512
    block_size = 1024
    padding_token = 256

    test_suite = _get_features_extraction_test_suite(
        beg_size=beg_size,
        mid_size=mid_size,
        end_size=end_size,
        block_size=block_size,
        padding_token=padding_token,
    )

    ref_features_extraction_tests = []

    for test_info, test_content in test_suite:
        s = Buffer(test_content)
        features_v2 = Magika._extract_features_from_seekable(
            s,
            beg_size,
            mid_size,
            end_size,
            padding_token,
            block_size,
            use_inputs_at_offsets=True,
        )

        test_case = {
            "test_info": asdict(test_info),
            "content": base64.b64encode(test_content).decode("ascii"),
            "features_v2": asdict(features_v2),
        }
        ref_features_extraction_tests.append(test_case)

    ref_features_extraction_tests_path = _get_features_extration_tests_path()
    ref_features_extraction_tests_path.parent.mkdir(parents=True, exist_ok=True)
    ref_features_extraction_tests_path.write_bytes(
        gzip.compress(json.dumps(ref_features_extraction_tests).encode("ascii"))
    )


def _get_tests_cases_from_reference() -> List:
    ref_features_extraction_tests_path = _get_features_extration_tests_path()

    tests_cases = json.loads(
        gzip.decompress(ref_features_extraction_tests_path.read_bytes())
    )
    return tests_cases  # type: ignore[no-any-return]


def _get_features_extraction_test_suite(
    beg_size: int, mid_size: int, end_size: int, block_size: int, padding_token: int
) -> List[Tuple[TestInfo, bytes]]:
    # for now we only support tests with beg_size == mid_size == end_size
    features_size = beg_size
    assert mid_size == features_size
    assert end_size == features_size

    ws_num_options = [
        0,
        1,
        10,
        features_size - 1,
        features_size,
        features_size + 1,
        block_size - 1,
        block_size,
        block_size + 1,
        2 * block_size - 1,
        2 * block_size,
        2 * block_size + 1,
        2 * block_size + features_size - 1,
        2 * block_size + features_size,
        2 * block_size + features_size + 1,
    ]

    content_size_options = list(ws_num_options)
    content_size_options.extend([10_000, 100_000])

    test_suite = []
    for core_content_size in content_size_options:
        for left_ws_num in ws_num_options:
            for right_ws_num in ws_num_options:
                test_info = TestInfo(
                    beg_size=beg_size,
                    mid_size=mid_size,
                    end_size=end_size,
                    block_size=block_size,
                    padding_token=padding_token,
                    core_content_size=core_content_size,
                    left_ws_num=left_ws_num,
                    right_ws_num=right_ws_num,
                )

                content = _generate_content(test_info)
                test_suite.append((test_info, content))

    def _gen_test_info_and_content(
        core_content_size: int, left_ws_num: int, right_ws_num: int
    ) -> Tuple[TestInfo, bytes]:
        test_info = TestInfo(
            beg_size=512,
            mid_size=512,
            end_size=512,
            block_size=4096,
            padding_token=256,
            core_content_size=core_content_size,
            left_ws_num=left_ws_num,
            right_ws_num=right_ws_num,
        )

        content = (
            b" " * left_ws_num
            + _generate_pattern(core_content_size)
            + b" " * right_ws_num
        )

        return test_info, content

    # add tests about 0xXXXX_0xYYYY features
    for offset_to_test in [0, 0x8000, 0x8800, 0x9000, 0x9800, 0x9900]:
        for extra_len in [0, 1, 7, 8, 9]:
            for left_ws_num in [0, 1]:
                test_info, content = _gen_test_info_and_content(
                    core_content_size=offset_to_test + extra_len,
                    left_ws_num=left_ws_num,
                    right_ws_num=0,
                )
                test_suite.append((test_info, content))

    return test_suite


def _get_features_extration_tests_path() -> Path:
    return get_tests_data_dir() / "features_extraction" / "reference.json.gz"


if __name__ == "__main__":
    test_features_extraction_v2(debug=True)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from magika.types.content_type_info import ContentTypeInfo  # noqa: F401
from magika.types.content_type_label import ContentTypeLabel  # noqa: F401
from magika.types.magika_error import MagikaError  # noqa: F401
from magika.types.magika_prediction import MagikaPrediction  # noqa: F401
from magika.types.magika_result import MagikaResult  # noqa: F401
from magika.types.model import (  # noqa: F401
    ModelConfig,
    ModelFeatures,
    ModelOutput,
)
from magika.types.overwrite_reason import OverwriteReason  # noqa: F401
from magika.types.prediction_mode import PredictionMode  # noqa: F401
from magika.types.status import Status  # noqa: F401

__all__ = [
    "ContentTypeInfo",
    "ContentTypeLabel",
    "MagikaError",
    "MagikaPrediction",
    "MagikaResult",
    "ModelConfig",
    "ModelFeatures",
    "ModelOutput",
    "OverwriteReason",
    "PredictionMode",
    "Status",
]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import io
from pathlib import Path
from typing import BinaryIO


class Seekable(abc.ABC):
    def __init__(self) -> None:
        self._size = -1

    @property
    def size(self) -> int:
        return self._size

    @abc.abstractmethod
    def read_at(self, offset: int, size: int) -> bytes:
        pass

    def close(self) -> None:
        pass


class File(Seekable):
    def __init__(self, path: Path) -> None:
        super().__init__()
        self._f = open(path, "rb")
        self._size = path.stat().st_size

    def read_at(self, offset: int, size: int) -> bytes:
        if size == 0:
            return b""

        assert offset + size <= self.size
        self._f.seek(offset, 0)  # whence = 0: start of the file
        return self._f.read(size)

    def close(self) -> None:
        self._f.close()


class Buffer(Seekable):
    def __init__(self, buffer: bytes) -> None:
        self._buffer = buffer
        self._size = len(buffer)

    def read_at(self, offset: int, size: int) -> bytes:
        if size == 0:
            return b""

        assert offset + size <= self.size
        return self._buffer[offset : offset + size]


class Stream(Seekable):
    def __init__(self, stream: BinaryIO) -> None:
        self._stream = stream
        stream.seek(0, io.SEEK_END)
        self._size = stream.tell()

    def read_at(self, offset: int, size: int) -> bytes:
        if size == 0:
            return b""

        assert offset + size <= self.size
        self._stream.seek(offset)
        return self._stream.read(size)
